#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
"""
Managed data workflows for Prescient
"""
# stdlib
import os
# package
from . import rts_gmlc


class ManagedWorkflow:
    def __init__(self, name, workspace_name):
        self._name = name
        self._workspace_name = workspace_name
        self._datasets = {}
        # TODO: create instance of DMF

    @property
    def name(self):
        return self._name

    @property
    def workspace_name(self):
        return self._workspace_name

    def get_dataset(self, type_, **kwargs):
        """Creates and returns a dataset of the specified type. If called more than once with the
        same type of dataset, then returns the previous value.
        """
        ds = self._datasets.get(type_, None)
        if ds is not None:
            return ds
        dsf = DatasetFactory(type_, workflow=self)
        ds = dsf.create(**kwargs)
        self._datasets[type_] = ds
        # TODO: register new dataset with DMF
        return ds


class Dataset:
    def __init__(self, name):
        self.name = name
        self._meta = {}

    @property
    def meta(self):
        return self._meta.copy()

    def add_meta(self, key, value):
        self._meta[key] = value

    def __str__(self):
        lines = [
            "Metadata",
            "--------"
        ]
        for key, value in self._meta.items():
            lines.append("%s:" % key)
            lines.append(str(value))
        return "\n".join(lines)


class DatasetFactory:
    def __init__(self, type_, workflow=None):
        self._wf = workflow
        try:
            self.create = self._get_factory_function(type_)
        except KeyError:
            raise KeyError("Cannot create dataset of type '%s'" % type_)

    @classmethod
    def _get_factory_function(cls, name):
        # This could be more dynamic..
        if name == "rts-gmlc":

            def download_fn(**kwargs):
                rts_gmlc_dir = rts_gmlc.download()
                dataset = Dataset(name)
                dataset.add_meta("directory", rts_gmlc_dir)
                dataset.add_meta("files", os.listdir(rts_gmlc_dir))
                return dataset

            return download_fn
        elif name == "null":
            def fn(**kwargs):
                return None  # XXX: or do we need a NullDataset subclass?
            return fn
        else:
            raise KeyError(name)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

from types import ModuleType

from idaes.apps.grid_integration import DoubleLoopCoordinator as _DLC
from idaes.apps.grid_integration.utils import convert_marginal_costs_to_actual_costs


class PrescientPluginModule(ModuleType):
    def __init__(self, get_configuration, register_plugins):
        self.get_configuration = get_configuration
        self.register_plugins = register_plugins


class DoubleLoopCoordinator(_DLC):

    def register_plugins(self, context, options, plugin_config):
        super().register_plugins(context, options, plugin_config)
        
        context.register_after_get_initial_actuals_model_for_sced_callback(
            self.update_static_params
        )
        context.register_after_get_initial_actuals_model_for_simulation_actuals_callback(
            self.update_static_params
        )
        context.register_after_get_initial_forecast_model_for_ruc_callback(
            self.update_static_params
        )

    @property
    def prescient_plugin_module(self):
        return PrescientPluginModule(self.get_configuration, self.register_plugins)

    def _update_static_params(self, gen_dict):
        
        is_thermal = (
            self.bidder.bidding_model_object.model_data.generator_type == "thermal"
        )
        is_renewable = (
            self.bidder.bidding_model_object.model_data.generator_type == "renewable"
        )
        for param, value in self.bidder.bidding_model_object.model_data:
            if param == "gen_name" or value is None:
                continue
            elif (
                param in gen_dict
                and isinstance(gen_dict[param], dict)
                and gen_dict[param]["data_type"] == "time_series"
            ):
                # don't touch time varying things;
                # presumably they be updated later
                continue
            elif param == "p_cost":
                if is_thermal:
                    curve_value = convert_marginal_costs_to_actual_costs(value)
                    gen_dict[param] = {
                        "data_type": "cost_curve",
                        "cost_curve_type": "piecewise",
                        "values": curve_value,
                    }
                elif is_renewable:
                    gen_dict[param] = value
                else:
                    raise NotImplementedError(
                        "generator_type must be either 'thermal' or 'renewable'"
                    )

            else:
                gen_dict[param] = value

    def update_static_params(self, options, instance):

        gen_name = self.bidder.bidding_model_object.model_data.gen_name
        gen_dict = instance.data["elements"]["generator"][gen_name]
        self._update_static_params(gen_dict)

    def pass_static_params_to_DA(self, *args, **kwargs):
        pass

    def pass_static_params_to_RT(self, *args, **kwargs):
        pass



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
"""
Basic tests for H2 property package
"""
import pytest

from pyomo.environ import ConcreteModel, value, SolverFactory, \
    TerminationCondition, SolverStatus

from dispatches.properties.h2_ideal_vap import configuration

from idaes.core import FlowsheetBlock
from idaes.models.properties.modular_properties.base.generic_property \
    import GenericParameterBlock
from idaes.core.solvers import get_solver


def test_h2_props():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = GenericParameterBlock(**configuration)

    m.fs.state = m.fs.props.build_state_block(
        m.fs.config.time, defined_state=True)

    # Fix state
    m.fs.state[0].flow_mol.fix(1)
    m.fs.state[0].mole_frac_comp.fix(1)
    m.fs.state[0].temperature.fix(300)
    m.fs.state[0].pressure.fix(101325)

    # Initialize state
    m.fs.state.initialize()

    solver = get_solver()
    results = solver.solve(m.fs)

    # Check for optimal solution
    assert results.solver.termination_condition == \
        TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    # Verify against NIST tables
    assert value(m.fs.state[0].cp_mol) == pytest.approx(28.85, rel=1e-2)
    assert value(m.fs.state[0].enth_mol) == pytest.approx(53.51, rel=1e-2)
    assert value(m.fs.state[0].entr_mol) == pytest.approx(130.9, rel=1e-2)
    assert (value(m.fs.state[0].gibbs_mol/m.fs.state[0].temperature) ==
            pytest.approx(-130.7, rel=1e-2))

    # Try another temeprature
    m.fs.state[0].temperature.fix(500)

    results = solver.solve(m.fs)

    # Check for optimal solution
    assert results.solver.termination_condition == \
        TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    assert value(m.fs.state[0].cp_mol) == pytest.approx(29.26, rel=1e-2)
    assert value(m.fs.state[0].enth_mol) == pytest.approx(5880, rel=1e-2)
    assert value(m.fs.state[0].entr_mol) == pytest.approx(145.7, rel=1e-2)
    assert (value(m.fs.state[0].gibbs_mol/m.fs.state[0].temperature) ==
            pytest.approx(-134.0, rel=1e-2))

    # Try another temeprature
    m.fs.state[0].temperature.fix(900)

    results = solver.solve(m.fs)

    # Check for optimal solution
    assert results.solver.termination_condition == \
        TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    assert value(m.fs.state[0].cp_mol) == pytest.approx(29.88, rel=1e-2)
    assert value(m.fs.state[0].enth_mol) == pytest.approx(17680, rel=1e-2)
    assert value(m.fs.state[0].entr_mol) == pytest.approx(163.1, rel=1e-2)
    assert (value(m.fs.state[0].gibbs_mol/m.fs.state[0].temperature) ==
            pytest.approx(-143.4, rel=1e-2))



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

"""
This script uses the multiperiod model for the integrated ultra-supercritical
power plant with energy storage and performs market analysis using the
pricetaker assumption. The electricity prices, LMP (locational marginal prices)
are assumed to not change. The prices used in this study are either obtained
from a synthetic database.
"""

__author__ = "Naresh Susarla and Soraya Rawlings"


from pyomo.environ import (Param, Objective,
                           Expression, SolverFactory, value, Constraint)
import numpy as np
from dispatches.case_studies.fossil_case.ultra_supercritical_plant.storage.\
    multiperiod_integrated_storage_usc import create_multiperiod_usc_model

# For plots
from matplotlib import pyplot as plt
import matplotlib
matplotlib.rc('font', size=24)
plt.rc('axes', titlesize=24)
scaling_obj = 1
scaling_factor = 1


def _get_lmp(number_hours):

    # Select lmp source data and scaling factor according to that
    use_rts_data = False
    use_mod_rts_data = True
    if use_rts_data:
        print('>>>>>> Using RTS lmp data')
        with open('rts_results_all_prices_base_case.npy', 'rb') as f:
            price = np.load(f)
    elif use_mod_rts_data:
        price = [22.9684, 21.1168, 20.4, 20.419,
                  20.419, 21.2877, 23.07, 25,
                  18.4634, 0, 0, 0,
                  0, 0, 0, 0,
                  19.0342, 23.07, 200, 200,
                  200, 200, 200, 200]
    else:
        print('>>>>>> Using NREL lmp data')
        price = np.load("nrel_scenario_average_hourly.npy")

    if use_rts_data:
        lmp = price[0:number_hours].tolist()
    elif use_mod_rts_data:
        lmp = price

    return lmp


def run_pricetaker_analysis(ndays, nweeks, tank_status, tank_min, tank_max):

    # Add number of days and hours per week
    number_hours = 24 * ndays
    n_time_points = nweeks * number_hours

    # Get LMP
    lmp = _get_lmp(number_hours)
    # Create the multiperiod model object. You can pass arguments to your
    # "process_model_func" for each time period using a dict of dicts as
    # shown here.  In this case, it is setting up empty dictionaries for
    # each time period.
    m = create_multiperiod_usc_model(
        n_time_points=n_time_points, pmin=None, pmax=None
    )

    # Retrieve active process blocks (i.e. time blocks)
    blks = [m.period[t] for t in m.set_period]
    
    m.periodic_variable_pair_costraint = Constraint(
        expr=(m.period[24].fs.salt_inventory_hot ==
             m.period[1].fs.previous_salt_inventory_hot))

    # Add lmp market data for each block
    count = 0
    for blk in blks:
        blk.lmp_signal = Param(default=0, mutable=True)
        blk.revenue = lmp[count]*blk.fs.net_power * scaling_factor
        blk.operating_cost = Expression(
            expr=(
                (blk.fs.operating_cost
                 + blk.fs.plant_fixed_operating_cost
                 + blk.fs.plant_variable_operating_cost) / (365 * 24)
            ) * scaling_factor
        )
        blk.cost = Expression(expr=-(blk.revenue - blk.operating_cost))
        count += 1

    m.obj = Objective(expr=sum([blk.cost for blk in blks]) * scaling_obj)

    # Initial state for salt tank for different scenarios
    # Tank initial scenarios:"hot_empty","hot_full","hot_half_full"
    if tank_status == "hot_empty":
        blks[0].fs.previous_salt_inventory_hot.fix(1103053.48)
        blks[0].fs.previous_salt_inventory_cold.fix(tank_max-1103053.48)
    elif tank_status == "half_full":
        blks[0].fs.previous_salt_inventory_hot.fix(tank_max/2)
        blks[0].fs.previous_salt_inventory_cold.fix(tank_max/2)
    elif tank_status == "hot_full":
        blks[0].fs.previous_salt_inventory_hot.fix(tank_max-tank_min)
        blks[0].fs.previous_salt_inventory_cold.fix(tank_min)
    else:
        print("Unrecognized scenario! Try hot_empty, hot_full, or half_full")

    blks[0].fs.previous_power.fix(447.66)

    # Plot results
    opt = SolverFactory('ipopt')
    hot_tank_level = []
    cold_tank_level = []
    net_power = []
    hxc_duty = []
    hxd_duty = []
    for week in range(nweeks):
        print()
        print(">>>>>> Solving for week {}: {} hours of operation in {} day(s) "
              .format(week + 1, number_hours, ndays))

        opt.solve(m, tee=True)

        hot_tank_level.append(
            [(value(blks[i].fs.salt_inventory_hot) / scaling_factor)*1e-3
             for i in range(n_time_points)])
        cold_tank_level.append(
            [(value(blks[i].fs.salt_inventory_cold) / scaling_factor)*1e-3
             for i in range(n_time_points)])
        net_power.append(
            [value(blks[i].fs.net_power)
             for i in range(n_time_points)])
        hxc_duty.append(
            [value(blks[i].fs.hxc.heat_duty[0])*1e-6
             for i in range(n_time_points)])
        hxd_duty.append(
            [value(blks[i].fs.hxd.heat_duty[0])*1e-6
             for i in range(n_time_points)])

    return (lmp, m, blks, hot_tank_level, cold_tank_level,
            net_power, hxc_duty, hxd_duty)


def print_results(m, blks):
    c = 0
    print('Objective: {:.4f}'.format(value(m.obj)))
    for blk in blks:
        print()
        print('Period {}'.format(c+1))
        print(' Net power: {:.4f}'.format(
            value(blks[c].fs.net_power)))
        print(' Plant Power Out: {:.4f}'.format(
            value(blks[c].fs.plant_power_out[0])))
        print(' ES Turbine Power: {:.4f}'.format(
            value(blks[c].fs.es_turbine.work_mechanical[0])*(-1e-6)))
        print(' Plant heat duty: {:.4f}'.format(
            value(blks[c].fs.plant_heat_duty[0])))
        print(' Cost ($): {:.4f}'.format(value(blks[c].cost) / scaling_factor))
        print(' Revenue ($): {:.4f}'
              .format(value(blks[c].revenue) / scaling_factor))
        print(' Operating cost ($): {:.4f}'
              .format(value(blks[c].operating_cost) / scaling_factor))
        print(' Specific Operating cost ($/MWh): {:.4f}'.format(
            (value(blks[c].operating_cost) / scaling_factor) /
            value(blks[c].fs.net_power)))
        print(' Cycle efficiency (%): {:.4f}'.format(
            value(blks[c].fs.cycle_efficiency)))
        print(' Boiler efficiency (%): {:.4f}'.format(
            value(blks[c].fs.boiler_eff) * 100))
        print(' Boiler heat duty: {:.4f}'.format(
            value(blks[c].fs.boiler.heat_duty[0]) * 1e-6))
        print(' Boiler flow mol (mol/s): {:.4f}'.format(
            value(blks[c].fs.boiler.outlet.flow_mol[0])))
        print(' Previous salt inventory (mton): {:.4f}'.format(
            (value(blks[c].fs.previous_salt_inventory_hot) /
             scaling_factor) * 1e-3))
        print(' Salt from HXC (mton): {:.4f}'.format(
            value(blks[c].fs.hxc.tube_outlet.flow_mass[0]) * 3600 * 1e-3))
        print(' Salt from HXD (mton): {:.4f}'.format(
            value(blks[c].fs.hxd.shell_outlet.flow_mass[0]) * 3600 * 1e-3))
        print(' HXC Duty (MW): {:.4f}'.format(
            value(blks[c].fs.hxc.heat_duty[0]) * 1e-6))
        print(' HXD Duty (MW): {:.4f}'.format(
            value(blks[c].fs.hxd.heat_duty[0]) * 1e-6))
        print(' Split fraction to HXC: {:.4f}'.format(
            value(blks[c].fs.ess_hp_split.split_fraction[0, "to_hxc"])))
        print(' Split fraction to HXD: {:.4f}'.format(
            value(blks[c].fs.ess_bfp_split.split_fraction[0, "to_hxd"])))
        print(' Salt flow HXC (kg/s): {:.4f}'.format(
            value(blks[c].fs.hxc.tube_outlet.flow_mass[0])))
        print(' Salt flow HXD (kg/s): {:.4f}'.format(
            value(blks[c].fs.hxd.shell_outlet.flow_mass[0])))
        print(' Steam flow HXC (mol/s): {:.4f}'.format(
            value(blks[c].fs.hxc.shell_outlet.flow_mol[0])))
        print(' Steam flow HXD (mol/s): {:.4f}'.format(
            value(blks[c].fs.hxd.tube_outlet.flow_mol[0])))
        print(' Delta T in HXC (kg): {:.4f}'.format(
            value(blks[c].fs.hxc.delta_temperature_in[0])))
        print(' Delta T out HXC (kg): {:.4f}'.format(
            value(blks[c].fs.hxc.delta_temperature_out[0])))
        print(' Delta T in HXD (kg): {:.4f}'.format(
            value(blks[c].fs.hxd.delta_temperature_in[0])))
        print(' Delta T out HXD (kg): {:.4f}'.format(
            value(blks[c].fs.hxd.delta_temperature_out[0])))
        c += 1


def plot_results(ndays, nweeks, lmp, m, blks, hot_tank_level, cold_tank_level,
                 net_power, hxc_duty, hxd_duty, tank_max):
    max_power = 436  # in MW
    max_power_storage = 30  # in MW
    max_power_total = max_power + max_power_storage
    min_storage_heat_duty = 10  # in MW
    max_storage_heat_duty = 200  # in MW

    n_time_points = 24 * ndays
    hours = np.arange(n_time_points * nweeks)
    lmp_array = np.asarray(lmp[0:n_time_points])
    hot_tank_array = np.asarray(hot_tank_level[0:nweeks]).flatten()
    cold_tank_array = np.asarray(cold_tank_level[0:nweeks]).flatten()

    # Convert array to list to include hot tank level at time zero
    hot_tank_array0 = (value(blks[0].fs.previous_salt_inventory_hot) /
                       scaling_factor) * 1e-3
    cold_tank_array0 = (value(blks[0].fs.previous_salt_inventory_cold) /
                        scaling_factor) * 1e-3
    hours_list = hours.tolist() + [n_time_points]
    hot_tank_list = [hot_tank_array0] + hot_tank_array.tolist()
    cold_tank_list = [cold_tank_array0] + cold_tank_array.tolist()

    font = {'size': 16}
    plt.rc('font', **font)
    fig1, ax1 = plt.subplots(figsize=(12, 8))

    color = ['r', 'b', 'tab:green', 'k', 'tab:orange']
    ax1.set_xlabel('Time Period (hr)')
    ax1.set_ylabel('Salt Tank Level (metric ton)',
                   color=color[3])
    ax1.spines["top"].set_visible(False)
    ax1.spines["right"].set_visible(False)
    ax1.grid(linestyle=':', which='both',
             color='gray', alpha=0.30)
    plt.axhline((tank_max / scaling_factor)*1e-3, ls=':', lw=1.75,
                color=color[4])
    plt.text(n_time_points / 2 - 1.5,
             (tank_max / scaling_factor) * 1e-3 + 100, 'max salt',
             color=color[4])
    ax1.step(  # [x + 1 for x in hours], hot_tank_array,
        hours_list, hot_tank_list,
        marker='^', ms=4, label='Hot Salt',
        lw=1, color=color[0])
    ax1.step(  # [x + 1 for x in hours], hot_tank_array,
        hours_list, cold_tank_list,
        marker='v', ms=4, label='Cold Salt',
        lw=1, color=color[1])
    ax1.legend(loc="center right", frameon=False)
    ax1.tick_params(axis='y')
    ax1.set_xticks(np.arange(0, n_time_points*nweeks + 1, step=2))

    ax2 = ax1.twinx()
    ax2.set_ylabel('LMP ($/MWh)',
                   color=color[2])
    ax2.step([x + 1 for x in hours], lmp_array,
             marker='o', ms=3, alpha=0.5,
             ls='-', lw=1,
             color=color[2])
    ax2.tick_params(axis='y',
                    labelcolor=color[2])
    plt.savefig('salt_tank_level_24h.png')

    font = {'size': 18}
    plt.rc('font', **font)

    power_array = np.asarray(net_power[0:nweeks]).flatten()
    # Convert array to list to include net power at time zero
    power_array0 = value(blks[0].fs.previous_power)
    power_list = [power_array0] + power_array.tolist()

    fig2, ax3 = plt.subplots(figsize=(12, 8))
    ax3.set_xlabel('Time Period (hr)')
    ax3.set_ylabel('Net Power Output (MW)',
                   color=color[1])
    ax3.spines["top"].set_visible(False)
    ax3.spines["right"].set_visible(False)
    ax3.grid(linestyle=':', which='both',
             color='gray', alpha=0.30)
    plt.text(n_time_points / 2 - 3, max_power - 5.5, 'max plant power',
             color=color[4])
    plt.text(n_time_points / 2 - 2.8, max_power_total + 1, 'max net power',
             color=color[4])
    plt.axhline(max_power, ls='-.', lw=1.75,
                color=color[4])
    plt.axhline(max_power_total, ls=':', lw=1.75,
                color=color[4])
    ax3.step(hours_list, power_list,
             marker='o', ms=4,
             lw=1, color=color[1])
    ax3.tick_params(axis='y',
                    labelcolor=color[1])
    ax3.set_xticks(np.arange(0, n_time_points * nweeks + 1, step=2))

    ax4 = ax3.twinx()
    ax4.set_ylabel('LMP ($/MWh)',
                   color=color[2])
    ax4.step([x + 1 for x in hours], lmp_array,
             marker='o', ms=3, alpha=0.5,
             ls='-', lw=1,
             color=color[2])
    ax4.tick_params(axis='y',
                    labelcolor=color[2])
    plt.savefig('plant_power_24h.png')

    zero_point = True
    hxc_array = np.asarray(hxc_duty[0:nweeks]).flatten()
    hxd_array = np.asarray(hxd_duty[0:nweeks]).flatten()
    hxc_duty0 = 0  # zero since the plant is not operating
    hxc_duty_list = [hxc_duty0] + hxc_array.tolist()
    hxd_duty0 = 0  # zero since the plant is not operating
    hxd_duty_list = [hxd_duty0] + hxd_array.tolist()

    fig3, ax5 = plt.subplots(figsize=(12, 8))
    ax5.set_xlabel('Time Period (hr)')
    ax5.set_ylabel('Storage Heat Duty (MW)',
                   color=color[3])
    ax5.spines["top"].set_visible(False)
    ax5.spines["right"].set_visible(False)
    ax5.grid(linestyle=':', which='both',
             color='gray', alpha=0.30)
    plt.text(n_time_points / 2 - 2.2, max_storage_heat_duty + 1, 'max storage',
             color=color[4])
    plt.text(n_time_points / 2 - 2, min_storage_heat_duty - 6.5, 'min storage',
             color=color[4])
    plt.axhline(max_storage_heat_duty, ls=':', lw=1.75,
                color=color[4])
    plt.axhline(min_storage_heat_duty, ls=':', lw=1.75,
                color=color[4])
    if zero_point:
        ax5.step(hours_list, hxc_duty_list,
                 marker='^', ms=4, label='Charge',
                 color=color[0])
        ax5.step(hours_list, hxd_duty_list,
                 marker='v', ms=4, label='Discharge',
                 color=color[1])
    else:
        ax5.step([x + 1 for x in hours], hxc_array,
                 marker='^', ms=4, lw=1,
                 label='Charge',
                 color=color[0])
        ax5.step([x + 1 for x in hours], hxd_array,
                 marker='v', ms=4, lw=1,
                 label='Discharge',
                 color=color[1])
    ax5.legend(loc="center right", frameon=False)
    ax5.tick_params(axis='y',
                    labelcolor=color[3])
    ax5.set_xticks(np.arange(0, n_time_points * nweeks + 1, step=2))

    ax6 = ax5.twinx()
    ax6.set_ylabel('LMP ($/MWh)',
                   color=color[2])
    ax6.step([x + 1 for x in hours], lmp_array,
             marker='o', ms=3, alpha=0.5,
             ls='-', color=color[2])
    ax6.tick_params(axis='y',
                    labelcolor=color[2])
    plt.savefig('heat_exchanger_duties_24h.png')

    plt.show()


if __name__ == "__main__":

    ndays = 1
    nweeks = 1

    tank_status = "hot_empty"
    tank_min = 1  # in kg
    tank_max = 6739292  # in kg

    (lmp, m, blks, hot_tank_level, cold_tank_level, net_power,
     hxc_duty, hxd_duty) = run_pricetaker_analysis(ndays, nweeks, tank_status,
                                                   tank_min, tank_max)

    print_results(m, blks)

    plot_results(ndays, nweeks, lmp, m, blks, hot_tank_level,
                 cold_tank_level, net_power, hxc_duty, hxd_duty, tank_max)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
from _pytest.config import Config


_MARKERS = {
    'unit': 'quick tests that do not require a solver, must run in < 2 s',
    'component': 'quick tests that may require a solver',
    'integration': 'long duration tests',
    'fails_in_ci': 'tests that are known to be failing (or flaky) when run in a CI environment'
}


def pytest_configure(config: Config):

    for name, descr in _MARKERS.items():
        config.addinivalue_line(
            'markers', f'{name}: {descr}'
        )



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

"""Test for integrated storage with ultrasupercritical power plant model

"""

__author__ = "Naresh Susarla"

import pytest

from pyomo.environ import value
from pyomo.util.check_units import assert_units_consistent

from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver

from dispatches.case_studies.fossil_case.ultra_supercritical_plant import (
    ultra_supercritical_powerplant as usc)
from dispatches.case_studies.fossil_case.ultra_supercritical_plant.storage \
     import integrated_storage_with_ultrasupercritical_power_plant as isp

@pytest.fixture(scope="module")
def model():

    # Build ultra-supercritical plant base model
    m = usc.build_plant_model()
    usc.initialize(m)
    # Create a flowsheet, add properties, unit models, and arcs
    m = isp.create_integrated_model(m, max_power=436)

    # Give all the required inputs to the model
    isp.set_model_input(m)

    # Add scaling factor
    isp.set_scaling_factors(m)

    return m


@pytest.mark.integration
def test_build(model):

    if degrees_of_freedom(model) != 0:
        raise Exception("Degrees of freedom after model build is not zero")

@pytest.mark.integration
def test_initialization(model):
    # Check that the cost correlations in charge model are initialized
    # properly and have 0 degrees of freedom
    isp.initialize(model)

    if degrees_of_freedom(model) != 0:
        raise Exception("Degrees of freedom after model initialize is not zero")


@pytest.mark.integration
def test_costing_method(model):
    # Add missing functions to complete the charge model (add bounds
    # and disjunctions)
    m = isp.build_costing(model)

    # Initialize with bounds
    isp.initialize_with_costing(m)
    isp.add_bounds(m)

    if degrees_of_freedom(m) != 0:
        raise Exception("Degrees of freedom after model initialize is not zero")

@pytest.mark.integration
def test_main_function():

    optarg = {"max_iter": 300}
    solver = get_solver('ipopt', optarg)


    model = isp.main(max_power=436)

    model = isp.model_analysis(
        model,
        solver,
        power=460,
        max_power=436,
        tank_scenario="hot_empty",
        fix_power=False)

    assert value(model.fs.revenue) == pytest.approx(9649.22, abs=1e-1)
    assert value(model.obj) == pytest.approx(5.386, abs=1e-1)
    assert value(model.fs.hxd.area) == pytest.approx(2204.88,
                                                     abs=1e-1)

# TODO: Add the unit consistency check once the PR #2395 is merged in Pyomo
# @pytest.mark.integration
# def test_unit_consistency(m):
#     assert_units_consistent(m)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
"""
Turbo-Generator Set for a Hydrogen turbine.

Compressor -> Stoichiometric Reactor -> Turbine
Author: Konor Frick
Date: April 2, 2021
Notes: it is noted that in this example the hydrogen is compressed along
with the air in the compressor as opposed to having a separate fuel
injection system. Noting this is a simplified version of the H2 turbine.
"""

import pytest
from pyomo.environ import ConcreteModel, SolverFactory, \
    value, TerminationCondition, SolverStatus
from idaes.core import FlowsheetBlock


from dispatches.properties.hturbine_ideal_vap import configuration
import dispatches.properties.h2_reaction as reaction_props
from dispatches.unit_models import HydrogenTurbine
from idaes.models.properties.modular_properties.base.generic_property \
    import GenericParameterBlock


from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver


def test_build():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    # Air Properties
    m.fs.properties1 = GenericParameterBlock(**configuration)

    m.fs.reaction_params = reaction_props.\
        H2ReactionParameterBlock(property_package=m.fs.properties1)

    # Adding H2 turbine model
    m.fs.h2_turbine = HydrogenTurbine(
        property_package=m.fs.properties1,
        reaction_package=m.fs.reaction_params,
    )

    # Check build
    assert hasattr(m.fs.h2_turbine, "compressor")
    assert hasattr(m.fs.h2_turbine, "stoic_reactor")
    assert hasattr(m.fs.h2_turbine, "turbine")

    assert degrees_of_freedom(m) == 13

    # Inlet Conditions********************************************************
    # ************************************************************************
    # Inlet Conditions of the inlet to the compressor.
    m.fs.h2_turbine.compressor.inlet.flow_mol[0].fix(4135.2)
    m.fs.h2_turbine.compressor.inlet.temperature[0].fix(288.15)
    m.fs.h2_turbine.compressor.inlet.pressure[0].fix(101325)

    m.fs.h2_turbine.compressor.inlet.mole_frac_comp[0, "oxygen"].fix(0.188)
    m.fs.h2_turbine.compressor.inlet.mole_frac_comp[0, "argon"].fix(0.003)
    m.fs.h2_turbine.compressor.inlet.mole_frac_comp[0, "nitrogen"].fix(0.702)
    m.fs.h2_turbine.compressor.inlet.mole_frac_comp[0, "water"].fix(0.022)
    m.fs.h2_turbine.compressor.inlet.mole_frac_comp[0, "hydrogen"].fix(0.085)

    m.fs.h2_turbine.compressor.deltaP.fix(2.401e6)
    m.fs.h2_turbine.compressor.efficiency_isentropic.fix(0.86)

    # Specify the Stoichiometric Conversion Rate of hydrogen
    # in the equation shown below
    # H2(g) + 1/2 O2(g) --> H2O(g) + energy
    # Complete Combustion
    m.fs.h2_turbine.stoic_reactor.conversion.fix(0.99)

    # Turbine Parameters
    m.fs.h2_turbine.turbine.deltaP.fix(-2.401e6)
    m.fs.h2_turbine.turbine.efficiency_isentropic.fix(0.89)

    assert degrees_of_freedom(m) == 0

    solver = get_solver()

    # Begin Initialization and solve for the system.
    m.fs.h2_turbine.initialize()

    results = solver.solve(m, tee=False)

    # Check for optimal solution
    assert results.solver.termination_condition == \
        TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    # Compressor
    assert value(m.fs.h2_turbine.compressor.outlet.temperature[0]) == \
        pytest.approx(763.25, rel=1e-1)

    # Stoichiometric Reactor
    assert value(m.fs.h2_turbine.stoic_reactor.
                 outlet.mole_frac_comp[0, 'hydrogen']) == \
        pytest.approx(0.00085, rel=1e-1)
    assert value(m.fs.h2_turbine.stoic_reactor.
                 outlet.mole_frac_comp[0, 'nitrogen']) == \
        pytest.approx(0.73285, rel=1e-1)
    assert value(m.fs.h2_turbine.stoic_reactor.
                 outlet.mole_frac_comp[0, 'oxygen']) == \
        pytest.approx(0.15232, rel=1e-1)
    assert value(m.fs.h2_turbine.stoic_reactor.
                 outlet.mole_frac_comp[0, 'water']) == \
        pytest.approx(0.11085, rel=1e-1)
    assert value(m.fs.h2_turbine.stoic_reactor.
                 outlet.mole_frac_comp[0, 'argon']) == \
        pytest.approx(0.0031318, rel=1e-1)

    # Turbine
    assert value(m.fs.h2_turbine.turbine.inlet.temperature[0]) == \
        pytest.approx(1426.3, rel=1e-1)
    assert value(m.fs.h2_turbine.turbine.outlet.temperature[0]) == \
        pytest.approx(726.44, rel=1e-1)

    assert m.fs.h2_turbine.inlet is m.fs.h2_turbine.compressor.inlet
    assert m.fs.h2_turbine.outlet is m.fs.h2_turbine.turbine.outlet

    m.fs.h2_turbine.report()



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
"""
Test for ultra supercritical power plant flowsheet
"""

__author__ = "Naresh Susarla"

from unittest.mock import patch, MagicMock
import runpy
import pytest
from pyomo.environ import TerminationCondition, value
from pyomo.util.check_units import assert_units_consistent

from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver

from dispatches.case_studies.fossil_case.ultra_supercritical_plant \
     import ultra_supercritical_powerplant as usc

solver = get_solver()


@pytest.fixture(scope="module")
def model():
    m = usc.build_plant_model()
    return m


@pytest.mark.integration
def test_initialize(model):
    # check that the model initialized properly and has 0 degrees of freedom
    usc.initialize(model)
    assert(degrees_of_freedom(model) == 0)


@pytest.mark.integration
def test_unit_consistency(model):
    assert_units_consistent(model)


@pytest.mark.integration
def test_main_function():
    # Build ultra supercriticla power plant model
    m = usc.build_plant_model()

    # Initialize the model (sequencial initialization and custom routines)
    usc.initialize(m)

    # Ensure after the model is initialized, the degrees of freedom = 0
    assert degrees_of_freedom(m) == 0

    # User can import the model from build_plant_model for analysis
    # A sample analysis function is called below
    m_result = usc.model_analysis(m, solver)

    # View results in a process flow diagram
    usc.view_result("pfd_usc_powerplant_result.svg", m_result)


@pytest.mark.integration
def test_usc_model(model):
    result = solver.solve(model, tee=False)
    assert result.solver.termination_condition == TerminationCondition.optimal
    assert (value(model.fs.plant_power_out[0]) ==
            pytest.approx(436.466,
                          abs=1e-2))  # Ref: Report/USDOE/FE-0400"
    assert (value(model.fs.constraint_bfp_power[0]) ==
            pytest.approx(0,
                          abs=1e-2))


@pytest.mark.integration
def test_change_power(model):
    model.fs.plant_power_out[0].fix(300)
    model.fs.boiler.inlet.flow_mol[0].unfix()
    result = solver.solve(model, tee=False)
    assert result.solver.termination_condition == TerminationCondition.optimal
    assert (value(model.fs.boiler.inlet.flow_mol[0]) ==
            pytest.approx(12474.473,
                          abs=1e-2))


@pytest.mark.integration
def test_change_pressure(model):
    model.fs.plant_power_out[0].unfix()
    model.fs.boiler.inlet.flow_mol[0].fix(17854)
    model.fs.boiler.outlet.pressure.fix(27e6)
    result = solver.solve(model, tee=False)
    assert result.solver.termination_condition == TerminationCondition.optimal
    assert (value(model.fs.plant_power_out[0]) ==
            pytest.approx(446.15,
                          abs=1e-2))
    assert (value(model.fs.plant_heat_duty[0]) ==
            pytest.approx(940.4,
                          abs=1e-2))



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

__author__ = "Konor Frick, Radhakrishna Tumbalam Gooty"

"""
This file contains a function to construct the generic nuclear flowsheet, and 
a function to fix the degrees of freedom and initialize.

To simulate the flowsheet, use:
# Create a concrete model
mdl = ConcreteModel()

# Build the nuclear flowsheet
build_ne_flowsheet(mdl)

# Fix the degrees of freedom and initialize the flowsheet
fix_dof_and_initialize(mdl)

# Ensure that the number of degrees of freedom is zero
print("Degrees of freedom: ", degrees_of_freedom(mdl))
assert degrees_of_freedom(mdl) == 0

# Simulate the entire flowsheet
res = get_solver().solve(mdl, tee=True)

see multiperiod_design_pricetaker.ipynb for more details
"""
# Pyomo imports
from pyomo.environ import (
    Constraint,
    ConcreteModel,
    TransformationFactory,
    units as pyunits,
)
from pyomo.network import Arc

# IDAES imports
from idaes.core import FlowsheetBlock
from idaes.models.unit_models import (
    Translator,
    Mixer,
    MomentumMixingType,
)
from idaes.models.properties.modular_properties.base.generic_property \
    import GenericParameterBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import propagate_state
import idaes.logger as idaeslog

# DISPATCHES imports
from dispatches.properties.h2_ideal_vap import configuration as h2_ideal_config
from dispatches.properties.hturbine_ideal_vap import configuration as hturbine_config
import dispatches.properties.h2_reaction as h2_reaction_props
from dispatches.unit_models import (
    PEM_Electrolyzer,
    ElectricalSplitter,
    SimpleHydrogenTank,
    HydrogenTurbine,
)


def build_ne_flowsheet(
    m=None,
    np_capacity=500,
    include_pem=True,
    include_tank=True,
    include_turbine=True,
    pem_capacity=None,
    tank_capacity=None,
    turbine_capacity=None,
):
    """Build the flowsheet for the nuclear case study

    This function builds the entire nuclear flowsheet by adding the
    required models and arcs connecting the models.

    Args:
        m: Pyomo `Block` or `ConcreteModel` (default: None)
        np_capacity: Capacity of the nuclear power plant [in MW] (default: 500)
        include_pem: If True, includes PEM electrolyzer (default: True)
        include_tank: If True, includes a simple hydrogen tank model (default: True)
        include_turbine: If True, includes the hydrogen turbine (default: True)
        pem_capacity: Capacity of the PEM electrolyzer [in MW] (default: None)
        tank_capacity: Capacity of the hydrogen tank [in kg] (default: None)
        turbine_capacity: Capacity of the hydrogen turbine [in MW] (default: None)

    Returns:
        Pyomo `Block` or `ConcreteModel` containing the flowsheet
    """
    if m is None:
        m = ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    # Load thermodynamic and reaction packages
    m.fs.h2ideal_props = GenericParameterBlock(**h2_ideal_config)
    hturbine_config["state_bounds"]["flow_mol"] = (0, 100, 1e5, pyunits.mol / pyunits.s)
    m.fs.h2turbine_props = GenericParameterBlock(**hturbine_config)
    m.fs.reaction_params = h2_reaction_props.H2ReactionParameterBlock(
        property_package=m.fs.h2turbine_props,
    )

    # Molecular weight of hydrogen
    mw_h2 = m.fs.h2ideal_props.config["components"]["hydrogen"]["parameter_data"]["mw"][0]

    # Add electrical splitter
    m.fs.np_power_split = ElectricalSplitter(
        num_outlets=2,
        outlet_list=["np_to_grid", "np_to_pem"],
        add_split_fraction_vars=True,
    )

    m.fs.np_power_split.electricity[0].fix(np_capacity * 1e3)  # Converting MW to kW

    if include_pem:
        # Add PEM electrolyzer
        m.fs.pem = PEM_Electrolyzer(property_package=m.fs.h2ideal_props)

        # Connect the electrical splitter and PEM
        m.fs.arc_np_to_pem = Arc(
            source=m.fs.np_power_split.np_to_pem_port,
            destination=m.fs.pem.electricity_in
        )

        if pem_capacity is not None:
            m.fs.pem.electricity_in.electricity.setub(pem_capacity * 1e3)  # Converting MW to kW

    else:
        include_tank = False
        include_turbine = False
        m.fs.np_power_split.split_fraction["np_to_pem", 0].fix(0)

    if include_tank:
        # Add hydrogen tank
        m.fs.h2_tank = SimpleHydrogenTank(property_package=m.fs.h2ideal_props)

        # Connect the pem electrolyzer and h2 tank
        m.fs.arc_pem_to_h2_tank = Arc(
            source=m.fs.pem.outlet,
            destination=m.fs.h2_tank.inlet
        )

        if tank_capacity is not None:
            m.fs.h2_tank.tank_holdup_previous.setub(tank_capacity / mw_h2)

    else:
        include_turbine = False

    if include_turbine:
        # Add translator block
        m.fs.translator = Translator(
            inlet_property_package=m.fs.h2ideal_props,
            outlet_property_package=m.fs.h2turbine_props,
        )

        # Add translator block constraints
        m.fs.translator.eq_flow_hydrogen = Constraint(
            expr=m.fs.translator.inlet.flow_mol[0] ==
                m.fs.translator.outlet.flow_mol[0])

        m.fs.translator.eq_temperature = Constraint(
            expr=m.fs.translator.inlet.temperature[0] ==
                m.fs.translator.outlet.temperature[0])

        m.fs.translator.eq_pressure = Constraint(
            expr=m.fs.translator.inlet.pressure[0] ==
                m.fs.translator.outlet.pressure[0])

        # Add mixer block
        # using minimize pressure for all inlets and outlet of the mixer
        # because pressure of inlets is already fixed in flowsheet,
        # using equality will over-constrain
        m.fs.mixer = Mixer(
            momentum_mixing_type=MomentumMixingType.minimize,
            property_package=m.fs.h2turbine_props,
            inlet_list=["air_feed", "hydrogen_feed"],
        )

        # Add hydrogen turbine
        m.fs.h2_turbine = HydrogenTurbine(
            property_package=m.fs.h2turbine_props,
            reaction_package=m.fs.reaction_params,
        )

        # Connect h2 tank and translator
        m.fs.arc_h2_tank_to_translator = Arc(
            source=m.fs.h2_tank.outlet_to_turbine,
            destination=m.fs.translator.inlet
        )

        # Connect translator and mixer
        m.fs.arc_translator_to_mixer = Arc(
            source=m.fs.translator.outlet,
            destination=m.fs.mixer.hydrogen_feed
        )

        # Connect mixer and h2 turbine
        m.fs.arc_mixer_to_h2_turbine = Arc(
            source=m.fs.mixer.outlet,
            destination=m.fs.h2_turbine.inlet
        )

        if turbine_capacity is not None:
            m.fs.h2_turbine.turbine_capacity = Constraint(
                expr=-m.fs.h2_turbine.work_mechanical[0] <= turbine_capacity * 1e6
            )  # Converting from MW to W

    TransformationFactory("network.expand_arcs").apply_to(m)

    return m


def fix_dof_and_initialize(
    m,
    pem_outlet_pressure=1.01325,
    pem_outlet_temperature=300,
    air_h2_ratio=10.76,
    compressor_dp=24.01,
    split_frac_grid=0.99,
    tank_holdup_previous=0,
    flow_mol_to_turbine=1,
    flow_mol_to_pipeline=1,
    outlvl=idaeslog.WARNING,
):
    """Fix degrees of freedom and initialize the flowsheet

    This function fixes the degrees of freedom of each unit and initializes the entire flowsheet.

    Args:
        m: Pyomo `Block` or `ConcreteModel` containing the flowsheet
        pem_outlet_pressure: Outlet pressure of hydrogen from PEM electrolyzer [in bar] (default: 1.01325)
        pem_outlet_temperature: Outlet temperature of hydrogen from PEM electrolyzer [in K] (default: 300)
        air_h2_ratio: Ratio of molar flowrate of air to molar flowrate of hydrogen
                      entering the hydrogen turbine (default: 10.76)
        compressor_dp: Pressure difference [in bar] between the outlet and the inlet of
                        the hydrogen turbine's compressor. The same pressure difference
                        is used for hydrogen turbine's turbine. (default: 24.01)
        split_frac_grid: Split fraction of electricity to the grid (default: 0.99)
        tank_holdup_previous: Initial holdup of the tank [in mol] (default: 0)
        flow_mol_to_turbine: Molar flowrate of hydrogen to turbine [in mol/s] (default: 1)
        flow_mol_to_pipeline: Molar flowrate of hydrogen to pipeline [in mol/s] (default: 1)
        outlvl: Logger (default: idaeslog.WARNING)
    """

    if not hasattr(m.fs, "pem"):
        m.fs.np_power_split.initialize(outlvl=outlvl)

        return

    # Fix the dof of the electrical splitter and initialize
    m.fs.np_power_split.split_fraction["np_to_grid", 0].fix(split_frac_grid)

    m.fs.np_power_split.initialize(outlvl=outlvl)

    # Fix the dof of the electrolyzer and initialize
    # Conversion of kW to mol/sec of H2 based on H-tec design of 54.517kW-hr/kg
    m.fs.pem.electricity_to_mol.fix(0.002527406)
    m.fs.pem.outlet.pressure.fix(pem_outlet_pressure * 1e5)
    m.fs.pem.outlet.temperature.fix(pem_outlet_temperature)

    propagate_state(m.fs.arc_np_to_pem)
    m.fs.pem.initialize(outlvl=outlvl)

    if not hasattr(m.fs, "h2_tank"):
        return

    # Fix the dof of the tank and initialize
    m.fs.h2_tank.dt.fix(3600)
    m.fs.h2_tank.tank_holdup_previous.fix(tank_holdup_previous)
    m.fs.h2_tank.outlet_to_turbine.flow_mol.fix(flow_mol_to_turbine)
    m.fs.h2_tank.outlet_to_pipeline.flow_mol.fix(flow_mol_to_pipeline)
    m.fs.h2_tank.outlet_to_turbine.mole_frac_comp[0, "hydrogen"].fix(1)
    m.fs.h2_tank.outlet_to_pipeline.mole_frac_comp[0, "hydrogen"].fix(1)

    propagate_state(m.fs.arc_pem_to_h2_tank)
    m.fs.h2_tank.initialize(outlvl=outlvl)

    if not hasattr(m.fs, "h2_turbine"):
        m.fs.h2_tank.outlet_to_turbine.flow_mol.fix(0)
        m.fs.h2_tank.initialize(outlvl=outlvl)
        return

    # Fix the dof of the translator block and initialize
    m.fs.translator.outlet.mole_frac_comp[0, "hydrogen"].fix(0.99)
    m.fs.translator.outlet.mole_frac_comp[0, "oxygen"].fix(0.01 / 4)
    m.fs.translator.outlet.mole_frac_comp[0, "argon"].fix(0.01 / 4)
    m.fs.translator.outlet.mole_frac_comp[0, "nitrogen"].fix(0.01 / 4)
    m.fs.translator.outlet.mole_frac_comp[0, "water"].fix(0.01 / 4)

    propagate_state(m.fs.arc_h2_tank_to_translator)
    m.fs.translator.initialize(outlvl=outlvl)

    # Fix the degrees of freedom of mixer and initialize
    m.fs.mixer.air_feed.flow_mol[0].fix(
        m.fs.h2_tank.outlet_to_turbine.flow_mol[0].value * air_h2_ratio
    )
    m.fs.mixer.air_feed.temperature[0].fix(pem_outlet_temperature)
    m.fs.mixer.air_feed.pressure[0].fix(pem_outlet_pressure * 1e5)
    m.fs.mixer.air_feed.mole_frac_comp[0, "oxygen"].fix(0.2054)
    m.fs.mixer.air_feed.mole_frac_comp[0, "argon"].fix(0.0032)
    m.fs.mixer.air_feed.mole_frac_comp[0, "nitrogen"].fix(0.7672)
    m.fs.mixer.air_feed.mole_frac_comp[0, "water"].fix(0.0240)
    m.fs.mixer.air_feed.mole_frac_comp[0, "hydrogen"].fix(2e-4)

    propagate_state(m.fs.arc_translator_to_mixer)
    m.fs.mixer.initialize(outlvl=outlvl)

    # Fix the degrees of freedom of H2 turbine and initialize
    m.fs.h2_turbine.compressor.deltaP.fix(compressor_dp * 1e5)
    m.fs.h2_turbine.compressor.efficiency_isentropic.fix(0.86)
    m.fs.h2_turbine.stoic_reactor.conversion.fix(0.99)
    m.fs.h2_turbine.turbine.deltaP.fix(-compressor_dp * 1e5)
    m.fs.h2_turbine.turbine.efficiency_isentropic.fix(0.89)

    propagate_state(m.fs.arc_mixer_to_h2_turbine)
    m.fs.h2_turbine.initialize(outlvl=outlvl)

    return



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

from .battery import BatteryStorage
from .elec_splitter import ElectricalSplitter
from .heat_exchanger_tube import ConcreteTubeSide
from .hydrogen_tank import HydrogenTank
from .hydrogen_tank_simplified import SimpleHydrogenTank
from .hydrogen_turbine_unit import HydrogenTurbine
from .pem_electrolyzer import PEM_Electrolyzer
from .wind_power import Wind_Power
from .concrete_tes import ConcreteTES



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################

__author__ = "Andres J Calderon, Naresh Susarla, Miguel Zamarripa, " \
             "Jaffer Ghouse, Radhakrishna Tumbalam Gooty"

import pandas as pd

# Import Pyomo libraries
from pyomo.environ import (
    ConcreteModel,
    Set,
    Var,
    RangeSet,
    ConstraintList,
    TransformationFactory,
    units as pyunits,
)
from pyomo.network import Arc

# Import IDAES libraries
from idaes.core import FlowsheetBlock, MaterialBalanceType
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import propagate_state
from idaes.models.unit_models import (
    HeatExchanger,
    PressureChanger,
    MomentumMixingType,
    Heater,
    Separator,
)
from idaes.models.unit_models.heat_exchanger import delta_temperature_underwood_callback

from idaes.models_extra.power_generation.unit_models.helm import (
    HelmSplitter,
    HelmTurbineStage,
    HelmMixer as Mixer,
    HelmIsentropicCompressor as WaterPump,
    HelmNtuCondenser as CondenserHelm,
)
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.models.properties import iapws95 

from dispatches.unit_models import ConcreteTES


"""
This is a simple power plant model for supercritical coal-fired power plant
This model uses some of the generic unit models in place of complex
power generation unit models such as multi-stage turbine and 0D FWH.
For example, this model uses multiple pressure changers to
model multi-stage turbine and 0D heat exchangers to model feed water heaters.
Some of the parameters in the model such as feed water heater areas,
overall heat transfer coefficient, turbine efficiencies at multiple stages
have all been estimated in order to closely match with the data in
NETL baseline report.
Additional assumptions include using Heater blocks to model main steam boiler,
reheater, and condenser. Also, the throttle valves at turbine inlet are
not included in this model.
The model has 3 degrees of freedom to be specified by the user. These are the
conditions for main steam: Pressure (Pa), Temperature (K), and Flow (mol/s).
The default values fixed in the model are from the NETL Baseline report rev4.

"""

CONC_TES_DATA = {
    "num_tubes": 10000,
    "num_segments": 20,
    "num_time_periods": 2,
    "tube_length": 64.9,
    "tube_diameter": 0.0105664,
    "face_area": 0.00847,
    "therm_cond_concrete": 1,
    "dens_mass_concrete": 2240,
    "cp_mass_concrete": 900,
    "init_temperature_concrete": [
        750, 732.631579, 715.2631579, 697.8947368, 680.5263158, 663.1578947,
        645.7894737, 628.4210526, 611.0526316, 593.6842105, 576.3157895, 558.9473684,
        541.5789474, 524.2105263, 506.8421053, 489.4736842, 472.1052632, 454.7368421,
        437.3684211, 420
    ],
    "flow_mol_charge": 0.00958 * 1000 / 18.01528,
    "inlet_pressure_charge": 19600000,
    "inlet_temperature_charge": 865,
    "flow_mol_discharge": 3 / 18.01528,
    "inlet_pressure_discharge": 8.5e5, #15 MPa
    "inlet_temperature_discharge": 355
}
DIS_IN_PRES = 8.5e5
DIS_IN_TEMP = 355


def build_scpc_flowsheet(m=None, include_concrete_tes=True, conc_tes_data=CONC_TES_DATA):
    """
    Build model for a supercritical pulverized coal-fired power plant

    Args:
        m: Pyomo `ConcreteModel` or `Block`
        include_concrete_tes: If True, includes concrete thermal energy storage

    Returns:
        m
    """

    if m is None:
        m = ConcreteModel()
    
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.include_concrete_tes = include_concrete_tes

    # Add thermodynamic package.
    m.fs.prop_water_mix = iapws95.Iapws95ParameterBlock(
        phase_presentation=iapws95.PhaseType.MIX,
    )
    m.fs.prop_water_mix.default_scaling_factor["pressure", None] = 1e-5
    m.fs.prop_water_mix.default_scaling_factor["flow_mol", None] = 1e-2

    # Add sets to the model
    m.fs.set_turbines = RangeSet(9, doc="Set of turbines")
    m.fs.set_t_splitters = RangeSet(8, doc="Set of splitters")
    m.fs.set_fwh = Set(initialize=[1, 2, 3, 4, 6, 7, 8], doc="Set of feed water heaters")
    m.fs.set_fwh_mix = Set(initialize=[1, 2, 3, 5, 6, 7], doc="Set of mixers associated with fwh")

    """
    TURBINE TRAIN: This part builds, turbine models, splitters models
    after each turbine, and hp_splitter models
    """
    # Add HP steam splitter
    m.fs.hp_splitter = HelmSplitter(property_package=m.fs.prop_water_mix)

    # Add turbine stages
    m.fs.turbine = HelmTurbineStage(m.fs.set_turbines, property_package=m.fs.prop_water_mix)

    # Add boiler feed water turbine
    m.fs.bfpt = HelmTurbineStage(property_package=m.fs.prop_water_mix)

    # Add splitters associated with turbines. Except splitter 4,
    # all splitters have 2 outlets. Splitter 4 has three outlets
    m.fs.t_splitter = HelmSplitter(
        m.fs.set_t_splitters,
        property_package=m.fs.prop_water_mix,
        num_outlets=2,
        # Set the number of outlets for splitter 4
        initialize={4: {
            "property_package": m.fs.prop_water_mix,
            "num_outlets": 3,}
        },
    )

    """
    BOILER AND FEED WATER HEATERS: This part builds models for boiler, reheater,
    and feed water heaters.

    Feed water heaters (FWHs) are declared as 0D heat exchangers
    Tube side is for feed water & Shell side is for steam condensing
    Pressure drop on both sides are accounted for by setting the respective
    outlet pressure based on the following assumptions:
        (1) Feed water side: A constant 4% pressure drop is assumed
              on the feedwater side for all FWHs. For this,
              the outlet pressure is set to 0.96 times the inlet pressure,
              on the feed water side for all FWHs
        (2) Steam condensing side: Going from high pressure to
              low pressure FWHs, the outlet pressure of
              the condensed steam in assumed to be 10% more than that
              of the pressure of steam extracted for the immediately
              next lower pressure feedwater heater.
              e.g. the outlet condensate pressure of FWH 'n'
              = 1.1 * pressure of steam extracted for FWH 'n-1'
              In case of FWH1 the FWH 'n-1' is used for Condenser,
              and in case of FWH6, FWH 'n-1' is for Deaerator. Here,
              the steam pressure for FWH 'n-1' is known because the
              pressure ratios for turbines are fixed.
    The condensing steam is assumed to leave the FWH as saturated liquid
    Thus, each FWH is accompanied by 3 constraints, 2 for pressure drop
    and 1 for the enthalpy.

    Boiler section is set up using two heater blocks, as following:
    1) For the main steam the heater block is named 'boiler'
    2) For the reheated steam the heater block is named 'reheater'
    """
    # Add boiler
    m.fs.boiler = Heater(
        property_package=m.fs.prop_water_mix, 
        has_pressure_change=True,
    )

    # Add reheater
    m.fs.reheater = Heater(
        property_package=m.fs.prop_water_mix,
        has_pressure_change=True,
    )

    # Outlet temperature of boiler is set to 866.15 K
    @m.fs.boiler.Constraint(m.fs.time)
    def boiler_temperature_constraint(blk, t):
        return blk.control_volume.properties_out[t].temperature == 866.15  # K

    # Outlet temperature of reheater is set to 866.15 K
    @m.fs.reheater.Constraint(m.fs.time)
    def reheater_temperature_constraint(blk, t):
        return blk.control_volume.properties_out[t].temperature == 866.15  # K

    # Add feed water heaters
    m.fs.fwh = HeatExchanger(
        m.fs.set_fwh,
        delta_temperature_callback=delta_temperature_underwood_callback,
        hot_side_name="shell",
        cold_side_name="tube",
        shell={
            "property_package": m.fs.prop_water_mix,
            "material_balance_type": MaterialBalanceType.componentTotal,
            "has_pressure_change": True,
        },
        tube={
            "property_package": m.fs.prop_water_mix,
            "material_balance_type": MaterialBalanceType.componentTotal,
            "has_pressure_change": True,
        },
    )

    # Data needed to set the pressure change on side 1 (shell) of fwh
    # FWH1: 0.5 is the pressure ratio for turbine #9 (see set_inputs)
    # FWH2: 0.64^2 is the pressure ratio for turbine #8 (see set_inputs)
    # FWH3: 0.64^2 is the pressure ratio for turbine #7 (see set_inputs)
    # FWH4: 0.64^2 is the pressure ratio for turbine #6 (see set_inputs)
    # FWH6: 0.79^6 is the pressure ratio for turbine #4 (see set_inputs)
    # FWH7: 0.79^4 is the pressure ratio for turbine #3 (see set_inputs)
    # FWH8: 0.8^2 is the pressure ratio for turbine #2 (see set_inputs)

    pressure_ratio_list = {1: 0.5,
                           2: 0.64 ** 2,
                           3: 0.64 ** 2,
                           4: 0.64 ** 2,
                           6: 0.79 ** 6,
                           7: 0.79 ** 4,
                           8: 0.8 ** 2}
    for i in m.fs.set_fwh:
        obj = m.fs.fwh[i]

        # Outlet enthalpy of condensate in an FWH must be same as that of saturated liquid
        @obj.Constraint(m.fs.time)
        def vapor_frac_constraint(blk, t):
            return (blk.hot_side.properties_out[t].enth_mol ==
                    blk.hot_side.properties_out[t].enth_mol_sat_phase['Liq'])

        # Setting a 4% pressure drop on the feedwater side (P_out = 0.96 * P_in)
        @obj.Constraint(m.fs.time)
        def s2_pdrop_constraint(blk, t):
            return (blk.cold_side.properties_out[t].pressure ==
                    0.96 * blk.cold_side.properties_in[t].pressure)

        # Setting the outlet pressure of condensate to be 10% more than that of
        # steam routed to condenser, as described in FWH description
        @obj.Constraint(m.fs.time)
        def s1_pdrop_constraint(blk, t):
            return (blk.hot_side.properties_out[t].pressure ==
                    1.1 * pressure_ratio_list[i] * blk.hot_side.properties_in[t].pressure)

    """
    FEED WATER MIXERS: This part adds mixers associated with feed water heaters
    """
    # Mixer 5 has an additional inlet from feedwater heater, and Mixer 7 has an
    # additional inlet from hx_pump
    m.fs.fwh_mix = Mixer(
        m.fs.set_fwh_mix,
        momentum_mixing_type=MomentumMixingType.none,
        inlet_list=["steam", "drain"],
        property_package=m.fs.prop_water_mix,
        initialize={
            # Add feedwater inlet to the deaerator
            5: {"momentum_mixing_type": MomentumMixingType.none,
                "inlet_list": ["steam", "drain", "feedwater"],
                "property_package": m.fs.prop_water_mix},
            # Add an inlet from hx_pump to mixer 7
            7: {"momentum_mixing_type": MomentumMixingType.none,
                "inlet_list": ["steam", "drain", "from_storage"],
                "property_package": m.fs.prop_water_mix},
            },
        )

    # Pressure constraint for all mixers
    for i in m.fs.set_fwh_mix:
        obj = m.fs.fwh_mix[i]

        if i == 5:  # Deaerator
            # Since the deaerator's (fwh_mix[5]) 'feedwater' inlet has the
            # minimum pressure, the outlet pressure must be the same as that
            # of the 'feedwater' inlet

            @obj.Constraint(m.fs.time)
            def mixer_pressure_constraint(blk, t):
                return blk.mixed_state[t].pressure == blk.feedwater_state[t].pressure

        else:
            # For the remaining mixers, the steam from the splitters has the minimum
            # pressure, so the outlet pressure must be the same as that of inlet steam

            @obj.Constraint(m.fs.time)
            def mixer_pressure_constraint(blk, t):
                return blk.mixed_state[t].pressure == blk.steam_state[t].pressure

    """
    CONDENSATION STAGE: This part adds models for all the unit operations 
    related to the heat removal process.
    """
    # Add condenser mixer
    # The inlet 'main' refers to the main steam coming from the turbine train
    # Inlet 'bfpt' refers to the steam coming from the bolier feed pump turbine
    # Inlet 'drain' refers to the condensed steam from the feed water heater 1
    # Inlet 'makeup' refers to the make up water
    m.fs.condenser_mix = Mixer(
        momentum_mixing_type=MomentumMixingType.none,
        inlet_list=["main", "bfpt", "drain", "makeup"],
        property_package=m.fs.prop_water_mix,
    )

    # The outlet pressure of condenser mixer is equal to the minimum pressure
    # Since the turbine (#9) outlet (or, mixer inlet 'main') pressure
    # has the minimum pressure, the following constraint sets the outlet
    # pressure of the condenser mixer to the pressure of the inlet 'main'
    @m.fs.condenser_mix.Constraint(m.fs.time)
    def mixer_pressure_constraint(blk, t):
        return blk.mixed_state[t].pressure == blk.main_state[t].pressure

    # Add condenser
    m.fs.condenser = CondenserHelm(
        shell={"has_pressure_change": False, "property_package": m.fs.prop_water_mix},
        tube={"has_pressure_change": False, "property_package": m.fs.prop_water_mix},
    )

    """
    PUMPS: This part adds models for all the pumps used in the flowsheet
    """
    # Add condensate pump (Pumps water from the condenser outlet)
    m.fs.cond_pump = WaterPump(property_package=m.fs.prop_water_mix)

    # Add boiler feed water pump (Pumps water between feed water heaters)
    m.fs.bfp = WaterPump(property_package=m.fs.prop_water_mix)

    # Add a splitter before bfp to extract water for the discharge
    m.fs.bfp_splitter = HelmSplitter(property_package=m.fs.prop_water_mix)

    if include_concrete_tes:
        append_tes_unit_models(m, conc_tes_data)

    else:
        # Concrete TES is not included, so close the ports
        # in hp_splitter, bfp_splitter, fwh_mix[7]
        m.fs.hp_splitter.split_fraction[:, "outlet_2"].fix(0)
        m.fs.bfp_splitter.split_fraction[:, "outlet_2"].fix(0)
        m.fs.fwh_mix[7].from_storage.flow_mol.fix(0)
        m.fs.fwh_mix[7].from_storage.pressure.fix(24235081.4)
        m.fs.fwh_mix[7].from_storage.enth_mol.fix(40000)

    """
    This part adds flowsheet-level constraints
    """
    # The following constraint sets the outlet pressure of steam extracted
    # for boiler feed water turbine to be same as that of condenser
    @m.fs.Constraint(m.fs.time)
    def constraint_bfpt_out_pressure(blk, t):
        return (blk.bfpt.control_volume.properties_out[t].pressure ==
                blk.condenser_mix.mixed_state[t].pressure)

    # The following constraint demands that the work done by the
    # boiler feed water pump is same as that of boiler feed water turbine
    # Essentially, this says that boiler feed water turbine produces just
    # enough power to meet the demand of boiler feed water pump
    @m.fs.Constraint(m.fs.time)
    def constraint_bfp_power(blk, t):
        return blk.bfp.control_volume.work[t] + blk.bfpt.control_volume.work[t] == 0

    # Declare net power output variable
    m.fs.net_power_output = Var(
        m.fs.time,
        initialize=620 * 1e6,
        doc="Net Power We out from the power plant",
        units=pyunits.W,
    )

    #   Constraint on Plant Power Output
    #   Plant Power Out = Turbine Power - Power required for HX Pump
    @m.fs.Constraint(m.fs.time)
    def production_cons(blk, t):
        return (sum(blk.turbine[j].work_mechanical[t] for j in m.fs.set_turbines)
                + m.fs.cond_pump.control_volume.work[t] == -m.fs.net_power_output[t])

    create_arcs(m)

    return m


def append_tes_unit_models(m, tes_data):
    """
    STORAGE: This function adds TES model and the models for auxiliary
    equipment required for the TES unit
    """

    m.fs.tes = ConcreteTES(
        model_data=tes_data,
        property_package=m.fs.prop_water_mix,
        operating_mode="combined",
    )

    m.fs.hp_splitter_to_tes = Arc(source=m.fs.hp_splitter.outlet_2,
                                  destination=m.fs.tes.inlet_charge)
    m.fs.tes_to_fwh_mix_7 = Arc(source=m.fs.tes.outlet_charge,
                                destination=m.fs.fwh_mix[7].from_storage)

    # Connect the outlet_2 of bfp_splitter to the cold side of tes
    inlet_enthalpy_discharge = iapws95.htpx(T=DIS_IN_TEMP * pyunits.K, P=DIS_IN_PRES * pyunits.Pa)
    m.fs.tes.inlet_discharge.pressure.fix(DIS_IN_PRES)
    m.fs.tes.inlet_discharge.enth_mol.fix(inlet_enthalpy_discharge)

    # @m.fs.Constraint(m.fs.time)
    # def discharge_flow_constraint(blk, t):
    #     return blk.tes.inlet_discharge.flow_mol[t] <= 0.2 * blk.bfp_splitter.inlet.flow_mol[t]

    m.fs.bfp_splitter.split_fraction[:, "outlet_2"].fix(0)

    """
    DISCHARGE TRAIN: This part adds the turbine for the discharge 
    process, and other associated equipment.
    """
    m.fs.discharge_turbine = HelmTurbineStage(property_package=m.fs.prop_water_mix)

    # To avoid condensation inside the turbine, we impose the following
    # constraint on the outlet
    @m.fs.discharge_turbine.Constraint(m.fs.time)
    def turbine_enthalpy_constraint(blk, t):
        # return (blk.control_volume.properties_out[t].enth_mol ==
        #         blk.control_volume.properties_out[t].enth_mol_sat_phase['Vap'])
        # return blk.control_volume.properties_out[t].temperature == 310
        return blk.control_volume.properties_out[t].pressure == 6644

    m.fs.tes_to_dis_turbine = Arc(source=m.fs.tes.outlet_discharge,
                                  destination=m.fs.discharge_turbine.inlet)

    return


def create_arcs(m):
    """
    TURBINE TRAIN: This part connects models in turbine train
    """
    # The following dictionary contains information on the destination
    # of outlet_2 stream from t_splitter.
    split_fwh_map = {1: ["fwh", 8],
                     2: ["fwh_mix", 7],
                     3: ["fwh_mix", 6],
                     4: ["fwh_mix", 5],  # Deaerator
                     5: ["fwh", 4],
                     6: ["fwh_mix", 3],
                     7: ["fwh_mix", 2],
                     8: ["fwh_mix", 1]}

    for i in m.fs.set_t_splitters:
        # Add arc connecting turbine and splitter
        setattr(m.fs, "turbine_" + str(i) + "_to_splitter_" + str(i),
                Arc(source=m.fs.turbine[i].outlet,
                    destination=m.fs.t_splitter[i].inlet))

        # Add arc connecting the splitter's outlet_1 to turbine/reheater
        if i == 2:
            # First outlet from splitter 2 goes to the reheater
            setattr(m.fs, "splitter_" + str(i) + "_to_reheater",
                    Arc(source=m.fs.t_splitter[i].outlet_1,
                        destination=m.fs.reheater.inlet))
        else:
            setattr(m.fs, "splitter_" + str(i) + "_to_turbine_" + str(i + 1),
                    Arc(source=m.fs.t_splitter[i].outlet_1,
                        destination=m.fs.turbine[i + 1].inlet))

        # Add arc connecting the splitter's outlet_2 to fwh/fwh_mix
        if split_fwh_map[i][0] == "fwh":
            setattr(m.fs, "splitter_" + str(i) + "_to_fwh_" + str(split_fwh_map[i][1]),
                    Arc(source=m.fs.t_splitter[i].outlet_2,
                        destination=m.fs.fwh[split_fwh_map[i][1]].shell_inlet))
        else:
            setattr(m.fs, "splitter_" + str(i) + "_to_fwh_mix_" + str(split_fwh_map[i][1]),
                    Arc(source=m.fs.t_splitter[i].outlet_2,
                        destination=m.fs.fwh_mix[split_fwh_map[i][1]].steam))

    # Connect the third outlet from splitter 4 to bfpt
    m.fs.splitter_4_to_bfpt = Arc(source=m.fs.t_splitter[4].outlet_3,
                                  destination=m.fs.bfpt.inlet)

    """
    BOILER: This part connects models related to boilers/feed water heaters
    """
    m.fs.boiler_to_hp_splitter = Arc(source=m.fs.boiler.outlet,
                                     destination=m.fs.hp_splitter.inlet)

    m.fs.hp_splitter_to_turbine_1 = Arc(source=m.fs.hp_splitter.outlet_1,
                                        destination=m.fs.turbine[1].inlet)

    m.fs.reheater_to_turbine_3 = Arc(source=m.fs.reheater.outlet,
                                     destination=m.fs.turbine[3].inlet)

    for i in m.fs.set_fwh_mix:  # List of fwh mixers
        # Add arcs to connect the outlet of fwh_mix to fwh/bfp
        if i == 5:
            # Outlet of Deaerator goes to bfp_splitter
            m.fs.fwh_mix_5_to_bfp_splitter = Arc(source=m.fs.fwh_mix[5].outlet,
                                                 destination=m.fs.bfp_splitter.inlet)

            # Outlet_1 of bfp_splitter goes to bfp
            m.fs.bfp_splitter_to_bfp = Arc(source=m.fs.bfp_splitter.outlet_1,
                                           destination=m.fs.bfp.inlet)

            # Boiler feed water pump to feed water heater 6
            m.fs.bfp_to_fwh_6 = Arc(source=m.fs.bfp.outlet,
                                    destination=m.fs.fwh[6].tube_inlet)

        else:
            # Outlet of each fwh_mix goes to the shell_inlet of fwh
            setattr(m.fs, "fwh_mix_" + str(i) + "_to_fwh_" + str(i),
                    Arc(source=m.fs.fwh_mix[i].outlet,
                        destination=m.fs.fwh[i].shell_inlet))

        # Add arcs connecting the drains and the mixers
        setattr(m.fs, "fwh_" + str(i + 1) + "_to_fwh_mix_" + str(i),
                Arc(source=m.fs.fwh[i + 1].shell_outlet,
                    destination=m.fs.fwh_mix[i].drain))

        # Add arcs connecting feed water outlet to feed water inlet
        if i != 5:  # Skip Deaerator
            setattr(m.fs, "fwh_" + str(i) + "_to_fwh_" + str(i + 1),
                    Arc(source=m.fs.fwh[i].tube_outlet,
                        destination=m.fs.fwh[i + 1].tube_inlet))

    m.fs.fwh_4_to_fwh_mix_5 = Arc(source=m.fs.fwh[4].tube_outlet,
                                  destination=m.fs.fwh_mix[5].feedwater)

    m.fs.fwh_8_to_boiler = Arc(source=m.fs.fwh[8].tube_outlet,
                               destination=m.fs.boiler.inlet)

    """
    CONDENSER: This part connects models related to the heat removal process
    """
    # Last turbine outlet to the mixer associated with the condenser
    m.fs.turbine_9_to_condenser_mix = Arc(source=m.fs.turbine[9].outlet,
                                          destination=m.fs.condenser_mix.main)

    # FWH to the mixer associated with condenser
    m.fs.fwh_1_to_condenser_mix = Arc(source=m.fs.fwh[1].shell_outlet,
                                      destination=m.fs.condenser_mix.drain)

    # Boiler feed water turbine to the mixer associated with condenser
    m.fs.bfpt_to_condenser_mix = Arc(source=m.fs.bfpt.outlet,
                                     destination=m.fs.condenser_mix.bfpt)

    # Mixer associated with condenser to condenser
    m.fs.condenser_mix_to_condenser = Arc(source=m.fs.condenser_mix.outlet,
                                          destination=m.fs.condenser.shell_inlet)

    # Condenser outlet to condenser pump
    m.fs.condenser_to_pump = Arc(source=m.fs.condenser.shell_outlet,
                                 destination=m.fs.cond_pump.inlet)

    # Condenser pump outlet to feed water heater
    m.fs.cond_pump_to_fwh_1 = Arc(source=m.fs.cond_pump.outlet,
                                  destination=m.fs.fwh[1].tube_inlet)

    TransformationFactory("network.expand_arcs").apply_to(m.fs)


def fix_dof_and_initialize(
    m, 
    outlvl=idaeslog.INFO, 
    hp_split_fraction=0.1,
    discharge_flow = 1,
):
    """
    Model inputs / fixed variable or parameter values
    assumed in this block, unless otherwise stated explicitly,
    are either assumed or estimated in order to match the results with
    known baseline scenario for supercritical steam cycle

    1. Each turbine has two dof: pressure ratio and isentropic efficiency
    2. Boiler feed water turbine has one dof: isentropic efficiency. The other
       dof is imposed as a constraint that requires the outlet pressure to be
       the same as the pressure of the condenser mixer outlet.
    3. t_splitters DO NOT have any dof. The split fraction is governed by the
       constraint that requires the outlet_1 of each fwh to be at saturated
       liquid state. However, since splitter 4 outlet does not go to a fwh, the
       above constraint is not applicable for it. So, we set the split fraction
       of the stream going to fwh_mix[5] i.e., deaerator. The other split fraction
       is governed by the constraint that -bfpt_power = bfp_power. In short, of the
       8 t_splitters, t_splitter[4] has one dof and the rest do not have any dof.
    4. hp_splitter and ip splitter have one dof: split fraction
    5. Boiler has two dof: outlet temperature and pressure. In addition to
       these two, we specify the inlet molar flowrate and remove one dof
       from condenser_mix
    6. Reheater has two dof: outlet temperature and pressure drop
    7. feed water heaters have two dof: area and heat transfer coefficient
    8. mixers associated with feed water heaters DO NOT have any dof
    9. Condenser mixer has three dof: state of the makeup. However, we will not
       fix the molar flowrate since we fixed an extra variable in boiler model
    10. Condenser has four dof: pressure and molar enthalpy of tube_inlet, area and
        heat transfer coefficient. The condenser model has a constraint that
        governs the molar flowrate of tube_inlet.
    11. Condenser pump has two dof: deltaP and isentropic efficiency
    12. Boiler feed water pump has two dof: outlet pressure and isentropic efficiency
    13. hx_pump has two dof: outlet pressure and isentropic efficiency
    14. Storage cooler has one dof. The other dof is imposed as a constraint
        on the outlet state (Outlet is at saturated liquid state).
    """

    main_steam_pressure = 24235081.4  # Pa
    boiler_inlet_flow_mol = 29111  # mol/s
    reheater_pressure_drop = -96526.64  # Pa, based on NETL Baseline report

    turbine_dof = {1: {"pr": 0.80 ** 5, "eta": 0.94},
                   2: {"pr": 0.80 ** 2, "eta": 0.94},
                   3: {"pr": 0.79 ** 4, "eta": 0.88},
                   4: {"pr": 0.79 ** 6, "eta": 0.88},
                   5: {"pr": 0.64 ** 2, "eta": 0.78},
                   6: {"pr": 0.64 ** 2, "eta": 0.78},
                   7: {"pr": 0.64 ** 2, "eta": 0.78},
                   8: {"pr": 0.64 ** 2, "eta": 0.78},
                   9: {"pr": 0.50, "eta": 0.78}}

    fwh_dof = {1: {"area": 400, "htc": 2000},
               2: {"area": 300, "htc": 2900},
               3: {"area": 200, "htc": 2900},
               4: {"area": 200, "htc": 2900},
               6: {"area": 600, "htc": 2900},
               7: {"area": 400, "htc": 2900},
               8: {"area": 400, "htc": 2900}}

    # Fix the degrees of freedom of turbines
    for i in turbine_dof:
        m.fs.turbine[i].ratioP.fix(turbine_dof[i]["pr"])
        m.fs.turbine[i].efficiency_isentropic.fix(turbine_dof[i]["eta"])

    # Boiler feed water turbine. The outlet pressure is set as a constraint
    m.fs.bfpt.efficiency_isentropic.fix(0.80)

    # Splitters
    m.fs.t_splitter[4].split_fraction[:, "outlet_2"].fix(0.050331)

    if m.fs.include_concrete_tes:
        m.fs.hp_splitter.split_fraction[:, "outlet_2"].fix(hp_split_fraction)
        m.fs.bfp_splitter.split_fraction[:, "outlet_2"].fix(0)

        # Fix degrees of freedom of the discharge turbine. we need to fix
        # only one dof for this turbine. The other dof is specified as a
        # constraint on the state of the outlet.
        m.fs.discharge_turbine.efficiency_isentropic.fix(0.75)

    # Fix the degrees of freedom of boiler
    m.fs.boiler.inlet.flow_mol.fix(boiler_inlet_flow_mol)
    m.fs.boiler.outlet.pressure.fix(main_steam_pressure)

    # Fix the degrees of freedom of reheater
    m.fs.reheater.deltaP.fix(reheater_pressure_drop)

    # Fix the degrees of freedom of feed water heaters
    for i in fwh_dof:
        m.fs.fwh[i].area.fix(fwh_dof[i]["area"])
        m.fs.fwh[i].overall_heat_transfer_coefficient.fix(fwh_dof[i]["htc"])

    # Make up stream to condenser
    m.fs.condenser_mix.makeup.flow_mol.value = 1.08002495835536E-12  # mol/s
    m.fs.condenser_mix.makeup.pressure.fix(103421.4)  # Pa
    m.fs.condenser_mix.makeup.enth_mol.fix(1131.69204)  # J/mol

    # Condenser
    m.fs.condenser.tube_inlet.pressure.fix(500000)
    m.fs.condenser.tube_inlet.enth_mol.fix(1800)
    m.fs.condenser.area.fix(34000)
    m.fs.condenser.overall_heat_transfer_coefficient.fix(3100)

    # Condenser pump
    m.fs.cond_pump.efficiency_isentropic.fix(0.80)
    m.fs.cond_pump.deltaP.fix(1e6)

    # Boiler feed water pump. Outlet pressure is assumed to be 15% more than the
    # desired main steam (Turbine Inlet) pressure to account for the pressure drop
    # across Feed water heaters and Boiler
    m.fs.bfp.efficiency_isentropic.fix(0.80)
    m.fs.bfp.outlet.pressure.fix(main_steam_pressure * 1.15)  # Pa

    """
    Initialization Begins
    """
    # Some numbers useful for initialization, but THESE ARE NOT DOF. We fix these
    # quantities at the beginning of the initialization process and unfix them
    # once the flowsheet has been initialized
    # FIXME: Had to change the value of fwh_mix_drain_state[1]["flow"]
    boiler_inlet_state = {"pressure": 24657896, "enthalpy": 20004}
    condenser_drain_state = {"flow": 1460, "pressure": 7308, "enthalpy": 2973}
    fwh_mix_drain_state = {
        1: {"flow": 1434 * 1.1, "pressure": 14617, "enthalpy": 3990},
        2: {"flow": 1136, "pressure": 35685, "enthalpy": 5462},
        3: {"flow": 788, "pressure": 87123, "enthalpy": 7160},
        5: {"flow": 6207, "pressure": 519291, "enthalpy": 11526},
        6: {"flow": 5299, "pressure": 2177587, "enthalpy": 16559},
        7: {"flow": 3730, "pressure": 5590711, "enthalpy": 21232},
    }

    # Fixing the split fractions of the splitters. Will be unfixed later
    m.fs.t_splitter[1].split_fraction[:, "outlet_2"].fix(0.12812)
    m.fs.t_splitter[2].split_fraction[:, "outlet_2"].fix(0.061824)
    m.fs.t_splitter[3].split_fraction[:, "outlet_2"].fix(0.03815)
    m.fs.t_splitter[4].split_fraction[:, "outlet_1"].fix(0.9019)
    m.fs.t_splitter[5].split_fraction[:, "outlet_2"].fix(0.0381443)
    m.fs.t_splitter[6].split_fraction[:, "outlet_2"].fix(0.017535)
    m.fs.t_splitter[7].split_fraction[:, "outlet_2"].fix(0.0154)
    m.fs.t_splitter[8].split_fraction[:, "outlet_2"].fix(0.00121)

    # Fix the boiler inlet state. Note flow_mol is a dof and is already fixed
    m.fs.boiler.inlet.pressure.fix(boiler_inlet_state["pressure"])
    m.fs.boiler.inlet.enth_mol.fix(boiler_inlet_state["enthalpy"])

    # Fix the condenser mixer's drain's state
    m.fs.condenser_mix.drain.flow_mol.fix(condenser_drain_state["flow"])
    m.fs.condenser_mix.drain.pressure.fix(condenser_drain_state["pressure"])
    m.fs.condenser_mix.drain.enth_mol.fix(condenser_drain_state["enthalpy"])

    # Fix the state of drain for all mixers
    for i in m.fs.set_fwh_mix:
        m.fs.fwh_mix[i].drain.flow_mol.fix(fwh_mix_drain_state[i]["flow"])
        m.fs.fwh_mix[i].drain.pressure.fix(fwh_mix_drain_state[i]["pressure"])
        m.fs.fwh_mix[i].drain.enth_mol.fix(fwh_mix_drain_state[i]["enthalpy"])

    # During initialization, we deactivate the constraint that the shell_outlet
    # should be a saturated liquid. This is because, we are fixing the split
    # fraction, so the chosen flowrate may not meet the requirement. We will
    # activate the constraint later after all the units have been initialized.
    for i in m.fs.set_fwh:
        m.fs.fwh[i].vapor_frac_constraint.deactivate()

    solver = get_solver(options={"tol": 1e-6,
                                 "max_iter": 300,
                                 "halt_on_ampl_error": "yes"})

    # Initializing the boiler
    m.fs.boiler.initialize(outlvl=outlvl, optarg=solver.options)

    # Initialize HP splitter
    propagate_state(m.fs.boiler_to_hp_splitter)
    m.fs.hp_splitter.initialize(outlvl=outlvl, optarg=solver.options)

    # Initialize the first turbine
    propagate_state(m.fs.hp_splitter_to_turbine_1)
    m.fs.turbine[1].initialize(outlvl=outlvl, optarg=solver.options)

    # Initialize the first splitter
    propagate_state(m.fs.turbine_1_to_splitter_1)
    m.fs.t_splitter[1].initialize(outlvl=outlvl, optarg=solver.options)

    # Initialize the second turbine
    propagate_state(m.fs.splitter_1_to_turbine_2)
    m.fs.turbine[2].initialize(outlvl=outlvl, optarg=solver.options)

    # Initialize the second splitter
    propagate_state(m.fs.turbine_2_to_splitter_2)
    m.fs.t_splitter[2].initialize(outlvl=outlvl, optarg=solver.options)

    # Initialize the reheater
    propagate_state(m.fs.splitter_2_to_reheater)
    m.fs.reheater.initialize(outlvl=outlvl, optarg=solver.options)

    # Initialize remaining turbines and splitters
    propagate_state(m.fs.reheater_to_turbine_3)
    for i in range(3, 10):
        m.fs.turbine[i].initialize(outlvl=outlvl, optarg=solver.options)

        if i < 9:  # Since there are only 8 splitters
            propagate_state(eval("m.fs.turbine_" + str(i) + "_to_splitter_" + str(i)))
            m.fs.t_splitter[i].initialize(outlvl=outlvl, optarg=solver.options)

            propagate_state(eval("m.fs.splitter_" + str(i) + "_to_turbine_" + str(i + 1)))

    # Initialize the boiler feed water turbine (Fixing the outlet pressure
    # to make it a square problem for initialization)
    propagate_state(m.fs.splitter_4_to_bfpt)
    m.fs.bfpt.outlet.pressure.fix(101325)
    m.fs.bfpt.initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.bfpt.outlet.pressure.unfix()

    # Initialize the condenser mixer
    propagate_state(m.fs.bfpt_to_condenser_mix)
    propagate_state(m.fs.turbine_9_to_condenser_mix)
    m.fs.condenser_mix.initialize(outlvl=outlvl, optarg=solver.options)

    # Initialize the condenser. We are fixing the molar flowrate of tube_inlet
    # to get a good initialization for feed water heaters
    propagate_state(m.fs.condenser_mix_to_condenser)
    m.fs.condenser.tube_inlet.flow_mol.fix(800000)
    m.fs.condenser.initialize(outlvl=outlvl, optarg=solver.options)
    m.fs.condenser.tube_inlet.flow_mol.unfix()

    # Initialize the condenser pump
    propagate_state(m.fs.condenser_to_pump)
    m.fs.cond_pump.initialize(outlvl=outlvl, optarg=solver.options)

    # Initialize feed water heaters, boiler feed water pump and mixers
    # associated with feed water heaters.
    # feed water heater 1
    propagate_state(m.fs.splitter_8_to_fwh_mix_1)
    m.fs.fwh_mix[1].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.fwh_mix_1_to_fwh_1)
    propagate_state(m.fs.cond_pump_to_fwh_1)
    m.fs.fwh[1].initialize(outlvl=outlvl, optarg=solver.options)

    # feed water heater 2 (Note: Drain state is fixed earlier)
    propagate_state(m.fs.splitter_7_to_fwh_mix_2)
    m.fs.fwh_mix[2].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.fwh_mix_2_to_fwh_2)
    propagate_state(m.fs.fwh_1_to_fwh_2)
    m.fs.fwh[2].initialize(outlvl=outlvl, optarg=solver.options)

    # feed water heater 3
    propagate_state(m.fs.splitter_6_to_fwh_mix_3)
    m.fs.fwh_mix[3].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.fwh_mix_3_to_fwh_3)
    propagate_state(m.fs.fwh_2_to_fwh_3)
    m.fs.fwh[3].initialize(outlvl=outlvl, optarg=solver.options)

    # feed water heater 4
    propagate_state(m.fs.splitter_5_to_fwh_4)
    propagate_state(m.fs.fwh_3_to_fwh_4)
    m.fs.fwh[4].initialize(outlvl=outlvl, optarg=solver.options)

    # Deaerator
    propagate_state(m.fs.splitter_4_to_fwh_mix_5)
    propagate_state(m.fs.fwh_4_to_fwh_mix_5)
    m.fs.fwh_mix[5].initialize(outlvl=outlvl, optarg=solver.options)

    # Initialize bfp_splitter
    propagate_state(m.fs.fwh_mix_5_to_bfp_splitter)
    m.fs.bfp_splitter.initialize(outlvl=outlvl, optarg=solver.options)

    # Initialize the TES system
    if m.fs.include_concrete_tes:
        propagate_state(m.fs.hp_splitter_to_tes)
        m.fs.tes.inlet_discharge.flow_mol.fix(discharge_flow)
        m.fs.tes.initialize(outlvl=outlvl, optarg=solver.options)

        propagate_state(m.fs.tes_to_fwh_mix_7)
        propagate_state(m.fs.tes_to_dis_turbine)
        m.fs.discharge_turbine.initialize(outlvl=outlvl, optarg=solver.options)

    # Boiler feed pump
    propagate_state(m.fs.bfp_splitter_to_bfp)
    m.fs.bfp.initialize(outlvl=outlvl, optarg=solver.options)

    # feed water heater 6
    propagate_state(m.fs.splitter_3_to_fwh_mix_6)
    m.fs.fwh_mix[6].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.fwh_mix_6_to_fwh_6)
    propagate_state(m.fs.bfp_to_fwh_6)
    m.fs.fwh[6].initialize(outlvl=outlvl, optarg=solver.options)

    # feed water heater 7
    propagate_state(m.fs.splitter_2_to_fwh_mix_7)
    m.fs.fwh_mix[7].initialize(outlvl=outlvl, optarg=solver.options)

    propagate_state(m.fs.fwh_mix_7_to_fwh_7)
    propagate_state(m.fs.fwh_6_to_fwh_7)
    m.fs.fwh[7].initialize(outlvl=outlvl, optarg=solver.options)

    # feed water heater 8
    propagate_state(m.fs.splitter_1_to_fwh_8)
    propagate_state(m.fs.fwh_7_to_fwh_8)
    m.fs.fwh[8].initialize(outlvl=outlvl, optarg=solver.options)

    # Unfix the values we fixed earlier for initialization
    # Unfix the inlet pressure and enthalpy of boiler. Flow_mol must remain fixed
    m.fs.boiler.inlet.pressure.unfix()
    m.fs.boiler.inlet.enth_mol.unfix()

    # Unfix the condenser mixer's drain's state
    m.fs.condenser_mix.drain.unfix()

    # Unfix the drain state of all the feed water heater mixers
    for i in m.fs.set_fwh_mix:
        m.fs.fwh_mix[i].drain.unfix()

    #  Unfix split fractions and activate vapor fraction constraints
    for i in m.fs.set_t_splitters:
        if i == 4:  # Split fraction to deaerator is fixed
            m.fs.t_splitter[i].split_fraction[:, "outlet_1"].unfix()
        else:
            m.fs.t_splitter[i].split_fraction[:, "outlet_2"].unfix()

        if i != 5:  # fwh[5] does not exist!
            m.fs.fwh[i].vapor_frac_constraint.activate()

    # for i in m.fs.set_turbines:
    #     m.fs.turbine[i].inlet.unfix()

    print(m.name, " Degrees of Freedom: ", degrees_of_freedom(m.fs))
    assert degrees_of_freedom(m.fs) == 0

    if outlvl == idaeslog.WARNING:
        res = solver.solve(m, tee=False)
    else:
        res = solver.solve(m, tee=True)
    
    print("Model Initialization = ", res.solver.termination_condition)

    print("*********************Model Initialized**************************")


def set_scaling_factors(m):
    """
    This function sets the missing scaling factors and any other scaling
    factors we want to include.
    """
    # for i in m.fs.set_turbines:
    #     iscale.set_scaling_factor(m.fs.turbine[i].control_volume.work, 1e-7)
    #
    # iscale.set_scaling_factor(m.fs.bfpt.control_volume.work, 1e-7)
    # iscale.set_scaling_factor(m.fs.boiler.control_volume.heat, 1e-9)
    # iscale.set_scaling_factor(m.fs.reheater.control_volume.heat, 1e-8)
    #
    # for i in m.fs.set_fwh:
    #     iscale.set_scaling_factor(m.fs.fwh[i].area, 1e-2)
    #     iscale.set_scaling_factor(m.fs.fwh[i].overall_heat_transfer_coefficient, 1e-3)
    #     iscale.set_scaling_factor(m.fs.fwh[i].shell.heat, 1e-7)
    #     iscale.set_scaling_factor(m.fs.fwh[i].tube.heat, 1e-7)
    #
    # iscale.set_scaling_factor(m.fs.condenser.shell.heat, 1e-9)
    # iscale.set_scaling_factor(m.fs.condenser.tube.heat, 1e-9)
    # iscale.set_scaling_factor(m.fs.cond_pump.control_volume.work, 1e-5)
    # iscale.set_scaling_factor(m.fs.bfp.control_volume.work, 1e-7)
    #
    # if m.fs.include_concrete_tes:
    #     m.fs.tes.set_default_scaling_factors()
    #     iscale.set_scaling_factor(m.fs.discharge_turbine.control_volume.work, 1)

    for i in m.fs.set_turbines:
        iscale.set_scaling_factor(m.fs.turbine[i].control_volume.work, 1e-6)

    iscale.set_scaling_factor(m.fs.bfpt.control_volume.work, 1e-6)
    iscale.set_scaling_factor(m.fs.boiler.control_volume.heat, 1e-7)
    iscale.set_scaling_factor(m.fs.reheater.control_volume.heat, 1e-7)

    for i in m.fs.set_fwh:
        iscale.set_scaling_factor(m.fs.fwh[i].area, 1e-2)
        iscale.set_scaling_factor(m.fs.fwh[i].overall_heat_transfer_coefficient, 1e-3)
        iscale.set_scaling_factor(m.fs.fwh[i].shell.heat, 1e-7)
        iscale.set_scaling_factor(m.fs.fwh[i].tube.heat, 1e-7)

    iscale.set_scaling_factor(m.fs.condenser.shell.heat, 1e-7)
    iscale.set_scaling_factor(m.fs.condenser.tube.heat, 1e-7)
    iscale.set_scaling_factor(m.fs.cond_pump.control_volume.work, 1e-6)
    iscale.set_scaling_factor(m.fs.bfp.control_volume.work, 1e-6)
    iscale.set_scaling_factor(m.fs.net_power_output, 1e-6)

    if m.fs.include_concrete_tes:
        m.fs.tes.set_default_scaling_factors()
        iscale.set_scaling_factor(m.fs.discharge_turbine.control_volume.work, 1e-4)

    iscale.calculate_scaling_factors(m)

    sf_enth_mol = m.fs.prop_water_mix.default_scaling_factor["enth_mol", None]
    sf_pressure = m.fs.prop_water_mix.default_scaling_factor["pressure", None]

    for i in m.fs.set_fwh:
        iscale.constraint_scaling_transform(
            m.fs.fwh[i].vapor_frac_constraint[0], sf_enth_mol)
        iscale.constraint_scaling_transform(
            m.fs.fwh[i].s1_pdrop_constraint[0], sf_pressure)
        iscale.constraint_scaling_transform(
            m.fs.fwh[i].s2_pdrop_constraint[0], sf_pressure)

    for i in m.fs.set_fwh_mix:
        iscale.constraint_scaling_transform(
            m.fs.fwh_mix[i].mixer_pressure_constraint[0], sf_pressure)

    iscale.constraint_scaling_transform(
        m.fs.condenser_mix.mixer_pressure_constraint[0], sf_pressure)
    iscale.constraint_scaling_transform(
        m.fs.constraint_bfpt_out_pressure[0], sf_pressure)
    iscale.constraint_scaling_transform(
        m.fs.constraint_bfp_power[0], 1e-6)
    iscale.constraint_scaling_transform(
        m.fs.production_cons[0], 1e-6)

    if m.fs.include_concrete_tes:
        iscale.constraint_scaling_transform(
            m.fs.discharge_turbine.turbine_enthalpy_constraint[0], sf_pressure)

        for t, con in m.fs.tes.charge_inlet_enth_mol_equality.items():
            iscale.constraint_scaling_transform(con, sf_enth_mol)

        for t, con in m.fs.tes.charge_inlet_pressure_equality.items():
            iscale.constraint_scaling_transform(con, sf_pressure)

        for t, con in m.fs.tes.charge_outlet_enth_mol_equality.items():
            iscale.constraint_scaling_transform(con, sf_enth_mol)

        for t, con in m.fs.tes.charge_outlet_pressure_equality.items():
            iscale.constraint_scaling_transform(con, sf_pressure)

        for t, con in m.fs.tes.discharge_inlet_enth_mol_equality.items():
            iscale.constraint_scaling_transform(con, sf_enth_mol)

        for t, con in m.fs.tes.discharge_inlet_pressure_equality.items():
            iscale.constraint_scaling_transform(con, sf_pressure)

        for t, con in m.fs.tes.discharge_outlet_enth_mol_equality.items():
            iscale.constraint_scaling_transform(con, sf_enth_mol)

        for t, con in m.fs.tes.discharge_outlet_pressure_equality.items():
            iscale.constraint_scaling_transform(con, sf_pressure)


def unfix_dof_for_optimization(m):
    """
    This function unfixes a few degrees of freedom for optimization
    For now, we will not unfix the design of the tes system. Also,
    we will not unfix the water flowrate to the boiler.
    Therefore, the boiler will always be maintained at its base load.
    The only decision we need to make is whether to produce power and
    sell it to the grid, and/or charge/discharge the TES. Since we have
    a separate discharge turbine, the total power produced could exceed
    the maximum rated capacity of the power plant.
    """
    if m.fs.include_concrete_tes:
        # Unfix the split fraction of the hp_splitter and specify bounds
        m.fs.hp_splitter.split_fraction[:, "outlet_2"].unfix()
        m.fs.hp_splitter.split_fraction[:, "outlet_2"].setlb(1e-5)
        m.fs.hp_splitter.split_fraction[:, "outlet_2"].setub(0.15)

        # Fixme: Currently, we have a separate inlet for the discharge.
        # so, the split fraction of bfp_splitter must be fixed. If this changed,
        # the the following lines need to be uncommented.
        # Unfix the split fraction of the bfp_spiltter
        # m.fs.bfp_splitter.split_fraction[:, "outlet_2"].unfix()
        # m.fs.bfp_splitter.split_fraction[:, "outlet_2"].setlb(1e-6)
        # m.fs.bfp_splitter.split_fraction[:, "outlet_2"].setub(0.05)

        m.fs.tes.inlet_discharge.flow_mol.unfix()
        m.fs.tes.inlet_discharge.flow_mol.setlb(1)
        m.fs.tes.inlet_discharge.flow_mol.setub(1650)

        # Unfix the initial temperature profile of the concrete block
        # Bounds on temperatures are set in the unit model
        m.fs.tes.period[1].concrete.init_temperature.unfix()

    else:
        # Unfix the boiler feed water flow
        m.fs.boiler.inlet.flow_mol.unfix()
        m.fs.boiler.inlet.flow_mol.setlb(14000)
        m.fs.boiler.inlet.flow_mol.setub(29112)


def generate_stream_table(m):
    blk = ConcreteModel()
    blk.thermo_params = iapws95.Iapws95ParameterBlock()
    blk.state = blk.thermo_params.build_state_block(
        defined_state=True,
        has_phase_equilibrium=True,
    )
    blk.state.flow_mol.fix(1)

    zz = {}
    for v in m.component_objects(Arc):
        zz[v.name] = {"flow_mol": v.ports[0].flow_mol[0].value,
                      "enth_mol": v.ports[0].enth_mol[0].value,
                      "pressure": v.ports[0].pressure[0].value}
        blk.state.pressure.fix(v.ports[0].pressure[0].value)
        blk.state.enth_mol.fix(v.ports[0].enth_mol[0].value)
        zz[v.name]["temperature"] = blk.state.temperature.expr()

    df = pd.DataFrame(zz)
    df.to_excel("SCPC_stream_table.xlsx")



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
"""
Basic tests for Compressed hydrogen gas (chg) tank model
Author: Naresh Susarla.
Date: May 27, 2021
"""

# Import Python libraries
import pytest

# Import Pyomo libraries
from pyomo.environ import (ConcreteModel,
                           SolverStatus,
                           Var,
                           value,
                           TerminationCondition)

# Import IDAES
from idaes.core.solvers import get_solver
from idaes.core import (FlowsheetBlock,
                        MomentumBalanceType)
from idaes.models.properties.modular_properties.base.generic_property \
    import GenericParameterBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
import idaes.core.util.scaling as iscale

# Import unit model and property package
from dispatches.unit_models import HydrogenTank
from dispatches.properties.h2_ideal_vap \
    import configuration

# Get default solver for testing
solver = get_solver()

@pytest.mark.unit
def test_config():
    # Create the ConcreteModel and the FlowSheetBlock
    m = ConcreteModel(name="H2TankModel")
    m.fs = FlowsheetBlock(dynamic=False)

    # Add hydrogen property parameter block
    m.fs.properties = GenericParameterBlock(**configuration)

    # Create an instance of the CHG tank
    m.fs.unit = HydrogenTank(dynamic=False, property_package=m.fs.properties,)

    # Check unit config arguments
    assert len(m.fs.unit.config) == 5

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.momentum_balance_type == \
        MomentumBalanceType.pressureTotal
    assert m.fs.unit.config.property_package is m.fs.properties

class TestH2IdealVap(object):
    @pytest.fixture(scope="class")
    def hydrogentank(self):
        # Create the ConcreteModel and the FlowSheetBlock
        m = ConcreteModel(name="H2TankModel")
        m.fs = FlowsheetBlock(dynamic=False)

        # Add hydrogen property parameter block
        m.fs.properties = GenericParameterBlock(**configuration)

        # Create an instance of the Hydrogen tank
        m.fs.unit = HydrogenTank(dynamic=False, property_package=m.fs.properties,)

        # Fix tank geometry
        m.fs.unit.tank_diameter.fix(0.1)
        m.fs.unit.tank_length.fix(0.3)

        # Fix initial state of tank
        m.fs.unit.previous_state[0].temperature.fix(300)
        m.fs.unit.previous_state[0].pressure.fix(1e5)

        # Fix inlet state
        m.fs.unit.control_volume.properties_in[0].flow_mol.fix(1)
        m.fs.unit.control_volume.properties_in[0].mole_frac_comp.fix(1)
        m.fs.unit.control_volume.properties_in[0].temperature.fix(300)
        m.fs.unit.control_volume.properties_in[0].pressure.fix(3e6)

        # Fix Duration of Operation (Time Step, 1hr = 3600s)
        m.fs.unit.dt[0].fix(3600)

        # Fix the outlet flow to zero for tank filling type operation
        m.fs.unit.control_volume.properties_out[0].flow_mol.fix(0)

        # Setting the bounds on the state variables
        m.fs.unit.control_volume.properties_in[0].pressure.setub(1e15)
        m.fs.unit.control_volume.properties_out[0].pressure.setub(1e15)
        m.fs.unit.previous_state[0].pressure.setub(1e15)

        return m

    @pytest.mark.unit
    def test_build(self, hydrogentank):
        assert hasattr(hydrogentank.fs.unit, "volume_cons")
        assert hasattr(hydrogentank.fs.unit, "control_volume")
        assert hasattr(hydrogentank.fs.unit, "previous_state")
        assert hasattr(hydrogentank.fs.unit, "material_balances")
        assert hasattr(hydrogentank.fs.unit, "energy_balances")
        assert hasattr(hydrogentank.fs.unit, "material_holdup_calculation")
        assert hasattr(hydrogentank.fs.unit, "material_holdup_integration")
        assert hasattr(hydrogentank.fs.unit, "energy_holdup_calculation")
        assert hasattr(hydrogentank.fs.unit, "energy_accumulation_equation")
        assert isinstance(hydrogentank.fs.unit.tank_diameter, Var)
        assert isinstance(hydrogentank.fs.unit.tank_length, Var)
        assert isinstance(hydrogentank.fs.unit.heat_duty, Var)
        assert isinstance(hydrogentank.fs.unit.previous_material_holdup, Var)
        assert isinstance(hydrogentank.fs.unit.previous_energy_holdup, Var)

    # Check degrees of freedom
    @pytest.mark.unit
    def test_dof(self, hydrogentank):
        assert degrees_of_freedom(hydrogentank) == 0

    # Check initialization
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, hydrogentank):
        initialization_tester(hydrogentank)

    # Check for optimal solution
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, hydrogentank):
        iscale.calculate_scaling_factors(hydrogentank)
        results = solver.solve(hydrogentank)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    # Tank filling scenario
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, hydrogentank):
        assert (value(hydrogentank.fs.unit.control_volume.\
                      properties_out[0].pressure)
                == pytest.approx(3820683416.393, rel=1e-1))
        assert (value(hydrogentank.fs.unit.control_volume.\
                      properties_out[0].temperature)
                == pytest.approx(300.749, rel=1e-1))
        assert (value(hydrogentank.fs.unit.control_volume.\
                      properties_out[0].dens_mol_phase["Vap"])
                == pytest.approx(1527927.5445, rel=1e-1))
        assert (value(hydrogentank.fs.unit.\
                      material_holdup[0, "Vap", "hydrogen"])
                == pytest.approx(3600.0945, rel=1e-1))

    # Try emptying scenario
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution2(self, hydrogentank):
        # Fix the outlet flow for tank emptying type operation
        hydrogentank.fs.unit.control_volume.properties_out[0].flow_mol.fix(0.9)
        results = solver.solve(hydrogentank)
        assert (results.solver.termination_condition ==
                TerminationCondition.optimal)

        assert (value(hydrogentank.fs.unit.control_volume.properties_out[0].pressure)
                == pytest.approx(381276651.957, rel=1e-1))
        assert (value(hydrogentank.fs.unit.control_volume.properties_out[0].temperature)
                == pytest.approx(300.055, rel=1e-1))
        assert (value(hydrogentank.fs.unit.control_volume.properties_out[0].\
                      dens_mol_phase["Vap"])
                == pytest.approx(152828.836, rel=1e-1))
        assert (value(hydrogentank.fs.unit.\
                      material_holdup[0, "Vap", "hydrogen"])
                == pytest.approx(360.0945, rel=1e-1))

    # Check report function
    @pytest.mark.unit
    def test_report(self, hydrogentank):
        hydrogentank.fs.unit.report()



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
"""
Property package for the hydrogen turbine burning of hydrogen and air to form water
"""

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import (Constraint,
                           exp,
                           Expression,
                           Param,
                           PositiveReals,
                           Set,
                           Var,
                           units as pyunits)

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        MaterialFlowBasis,
                        ReactionParameterBlock,
                        ReactionBlockDataBase,
                        ReactionBlockBase)
from idaes.core.util.misc import add_object_reference

# Set up logger
_log = logging.getLogger(__name__)


@declare_process_block_class("H2ReactionParameterBlock")
class H2ReactionParameterData(ReactionParameterBlock):
    """
    Property Parameter Block Class

    Contains parameters and indexing sets associated with properties for
    superheated steam.

    """
    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(H2ReactionParameterData, self).build()

        self._reaction_block_class = H2ReactionBlock

        # List of valid phases in property package
        self.phase_list = Set(initialize=['Vap'])

        # Component list - a list of component identifiers
        self.component_list = Set(initialize=['argon',
                                              'oxygen',
                                              'nitrogen',
                                              'hydrogen',
                                              'water'])

        # Reaction Index
        self.rate_reaction_idx = Set(initialize=["R1"])

        # Reaction Stoichiometry
        self.rate_reaction_stoichiometry = {("R1", "Vap", "argon"): 0,
                                            ("R1", "Vap", "oxygen"): -1,
                                            ("R1", "Vap", "nitrogen"): 0,
                                            ("R1", "Vap", "hydrogen"): -2,
                                            ("R1", "Vap", "water"): 2,
                                            ("R1", "Liq", "argon"): 0,
                                            ("R1", "Liq", "oxygen"): 0,
                                            ("R1", "Liq", "nitrogen"): 0,
                                            ("R1", "Liq", "hydrogen"): 0,
                                            ("R1", "Liq", "water"): 0}

        # Heat of Reaction
        dh_rxn_dict = {"R1": -4.8366e5}
        self.dh_rxn = Param(self.rate_reaction_idx,
                            initialize=dh_rxn_dict,
                            units=pyunits.J / pyunits.mol,
                            doc="Heat of reaction [J/mol]")

        # Gas Constant
        self.gas_const = Param(within=PositiveReals,
                               mutable=False,
                               default=8.314,
                               doc='Gas Constant [J/mol.K]')

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({
                'k_rxn': {'method': '_rate_constant', 'units': 'm^3/mol.s'},
                'reaction_rate': {'method': "_rxn_rate", 'units': 'mol/m^3.s'}
                })
        obj.add_default_units({'time': pyunits.s,
                               'length': pyunits.m,
                               'mass': pyunits.kg,
                               'amount': pyunits.mol,
                               'temperature': pyunits.K})


class ReactionBlock(ReactionBlockBase):
    """
    This Class contains methods which should be applied to Reaction Blocks as a
    whole, rather than individual elements of indexed Reaction Blocks.
    """
    def initialize(blk, outlvl=0, **kwargs):
        '''
        Initialization routine for reaction package.

        Keyword Arguments:
            outlvl : sets output level of initialization routine

                     * 0 = no output (default)
                     * 1 = report after each step

        Returns:
            None
        '''
        if outlvl > 0:
            _log.info('{} Initialization Complete.'.format(blk.name))


@declare_process_block_class("H2ReactionBlock",
                             block_class=ReactionBlock)
class H2ReactionBlockData(ReactionBlockDataBase):
    """
    An example reaction package for saponification of ethyl acetate
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(H2ReactionBlockData, self).build()

        # Heat of reaction - no _ref as this is the actual property
        add_object_reference(
                self,
                "dh_rxn",
                self.config.parameters.dh_rxn)

    def get_reaction_rate_basis(b):
        return MaterialFlowBasis.molar



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis Platform
# to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES), and is
# copyright (c) 2020-2023 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#################################################################################
"""
Tests for dispatches.workflow.workflow
"""
import pytest

from dispatches.workflow import ManagedWorkflow
from dispatches.workflow.workflow import Dataset, DatasetFactory


def test_managed_workflow():
    wf = ManagedWorkflow("hello", "world")
    assert wf.name == "hello"
    assert wf.workspace_name == "world"


def test_dataset():
    ds = Dataset("hello")
    assert ds.name == "hello"
    assert ds.meta == {}


def test_dataset_factory():
    df = DatasetFactory("null")
    assert df.create(hello="ignored") is None
    # with unknown type, raises KeyError
    pytest.raises(KeyError, DatasetFactory, "?")
